"""Module with covariance functions for kernels"""

import autograd.numpy as np


class Kernel:
    """Base kernel class. Shared functions"""
    
    def get_var(self, name, hyp_vec, key_dict, hyp_dict):
        if name in key_dict.keys():
            return hyp_vec[key_dict[name]]
        return hyp_dict.__dict__[name]
    
    def covariance_partials(
        self, x1_pp, x2_pp, tau, hyp_dict, dd_dime=None
    ):
        """Main function - covariance between two points"""
        raise NotImplementedError

    def log_prob(self, hyp_vec, xx, yy, key_dict, hyp_dict):
        """Log probability for kernel with hypereparameter
        settings from hyp_vec"""
        nn = np.size(yy)
        if self.name in ['Combination', 'Temporal']:
            try:
                sigma_2 = hyp_vec[key_dict[0]["sigma_2"]]
            except:
                sigma_2 = hyp_vec[key_dict[0][0]["sigma_2"]]
        else:
            sigma_2 = hyp_vec[key_dict["sigma_2"]]
        Ky = self.base_multi(xx, hyp_vec, key_dict, hyp_dict) + \
                sigma_2 * np.eye(nn) \
                + 0.00001 * np.eye(nn) # for numerical stability
        Ky_inv = np.linalg.inv(Ky)
        sign, logdet = np.linalg.slogdet(Ky)

        return -(
            -1 / 2 * np.matmul(yy.T, np.matmul(Ky_inv, yy))
            - 1 / 2 * sign * logdet
            - nn / 2 * np.log(2 * np.pi)
        )
    
    def log_prob_cholesky(self, hyp_vec, xx, yy, key_dict, hyp_dict):
        """Log probability for kernel with hypereparameter
        settings from hyp_vec"""
        nn = np.size(yy)
        if self.name in ['Combination', 'Temporal']:
            try:
                sigma_2 = hyp_vec[key_dict[0]["sigma_2"]]
            except:
                sigma_2 = hyp_vec[key_dict[0][0]["sigma_2"]]
        else:
            sigma_2 = hyp_vec[key_dict["sigma_2"]]
        Ky = self.base_multi(xx, hyp_vec, key_dict, hyp_dict) + \
                sigma_2 * np.eye(nn) \
                + 0.00001 * np.eye(nn) # for numerical stability
        
        L = np.linalg.cholesky(Ky)

        alpha = np.linalg.solve(L.T, np.linalg.solve(L, yy))

        return -(
            -1 / 2 * np.matmul(yy.T, alpha)
            - np.sum(np.log(np.diag(L)))
            - nn / 2 * np.log(2 * np.pi)
        )

    def zz_mean_cov_xx(self, xx, xx_train, yy_train, hyp_dict):
        """Get mean and covariance for z"""
        nn, D = np.shape(xx_train)
        num_upper = int(D * (D - 1) / 2)

        yy_ext = np.append(yy_train, [0] * (D + num_upper))
        K_sev = self.k_sev(xx_train, xx, hyp_dict)
        K_z = K_sev[: D + 1, : D + 1]
        K_c = K_sev[D + 1 :, D + 1 :]
        K_zc = K_sev[: 1 + D, 1 + D :]
        K_cz = K_sev[1 + D :, : 1 + D]

        noise_eye = np.zeros((nn + D + num_upper, nn + D + num_upper))
        for ii in range(nn):
            noise_eye[ii, ii] = hyp_dict.sigma_2
        A = np.matmul(
            K_zc,
            np.linalg.inv(
                K_c + noise_eye 
            ),
        )

        mean = np.array(np.matmul(A, yy_ext), ndmin=2).T
        cov = K_z - np.matmul(A, K_cz)
        return mean, cov

    def kmm_extended_gp(self, domain, hyp_dict, m_deg=0):
        """Get covariance between test points."""
        assert m_deg == 0
        return self.base_multi(domain, 
                               hyp_dict.array,
                               hyp_dict.key_dict,
                               hyp_dict)
        

    def dd_dims_helper(self, D):
        """Build dd_dims"""
        collector = []
        for aa in range(D - 1):
            for bb in range(aa + 1, D):
                collector.append([aa, bb])
        return collector

    def k_sev_helper(self, xx, xx_opt, ret_dict=False):
        """Calculate required values in k_sev and k_cross_sev"""
        nn, D = np.shape(xx)
        num_upper = int(D * (D - 1) / 2)
        vals = (
            [xx_opt] * (D + 1)
            + list(xx)
            + [xx_opt] * (D + num_upper)
        )
        n_ext = nn + 1 + 2 * D + num_upper
        degrees = (
            [0] + [2] * D + [0] * nn + [1] * D + [2] * num_upper
        )
        dd_dims = (
            [[]]
            + [[aa] * 2 for aa in range(D)]
            + [[]] * nn
            + [[aa] for aa in range(D)]
            + self.dd_dims_helper(D)
        )
        ksh = {'n_ext': n_ext,
               'degrees': degrees,
               'vals': vals,
               'dd_dims': dd_dims}
        if ret_dict:
            return ksh
        return n_ext, degrees, vals, dd_dims

    def k_sev(self, xx, xx_opt, hyp_dict):
        """Training covariance with pseudosamples"""
        n_ext, degrees, vals, dd_dims = self.k_sev_helper(
            xx, xx_opt
        )

        Knn = np.zeros((n_ext, n_ext))
        for ii in range(n_ext):
            for jj in range(n_ext):
                Knn[ii, jj] = self.covariance_partials(
                    degrees[ii],
                    degrees[jj],
                    vals[ii] - vals[jj],
                    hyp_dict,
                    dd_dims[ii] + dd_dims[jj]
                )
        return Knn

    def k_f_sev(self, xx_prod, xx_opt, hyp_dict):
        """Test covariance with pseudosamples"""
        return self.kmm_extended_gp(
            np.array([xx_prod, xx_opt]), hyp_dict, m_deg=0
        )

    def k_cross_sev(
        self, xx, xx_opt, xx_prod, hyp_dict, reverted=False
    ):
        """Training-test covariance with pseudosamples"""
        domain = np.array([xx_prod, xx_opt])
        mm = 2

        n_ext, degrees, vals, dd_dims = self.k_sev_helper(
            xx, xx_opt
        )

        return self.k_cross(
            domain,
            vals,
            n_ext,
            mm,
            degrees,
            dd_dims,
            hyp_dict,
            reverted,
        )
    
    def k_cross_sev_faster(
        self, xx, xx_opt, xx_prod, hyp_dict, ksh, 
        xx_opt_cross, reverted=False
    ):
        """Training-test covariance with pseudosamples"""

        xx_prod_cross = self.k_cross(
            np.array([xx_prod]),
            ksh['vals'], 
            ksh['n_ext'],
            1,
            ksh['degrees'],
            ksh['dd_dims'],
            hyp_dict,
            reverted,
        )
        return np.hstack([xx_prod_cross, xx_opt_cross])

    def k_cross(
        self,
        domain,
        vals,
        n_ext,
        mm,
        degrees,
        dd_dims,
        hyp_dict,
        reverted=False,
    ):
        """Training-test covariance
        For variance between one test point and training points use
        domain=x_test, vals=X_train, n_ext=N, mm=1, 
        degrees=np.zeros(N), dd_dims=[[]]*N
        where N is the number of training poitns
        """
        Knm = np.zeros((n_ext, mm))
        for ii in range(n_ext):
            for jj in range(mm):
                if not reverted:
                    Knm[ii, jj] = self.covariance_partials(
                        degrees[ii],
                        0,
                        vals[ii] - domain[jj],
                        hyp_dict,
                        dd_dims[ii] + [],
                    )
                else:
                    Knm[ii, jj] = self.covariance_partials(
                        0,
                        degrees[ii],
                        domain[jj] - vals[ii],
                        hyp_dict,
                        [] + dd_dims[ii],
                    )
        return Knm

    def cov_prior(self, x_prod, X, hyp_dict):
        """ sample prior covariance
        Used in pes calculation"""
        N, D = np.shape(X)
        k_prod = self.covariance_partials(
            0, 0, np.zeros(D), hyp_dict
        )
        K_n = self.kmm_extended_gp(X, hyp_dict, m_deg=0)
        k_n_prod = self.k_cross(
            np.array([x_prod]),
            X,
            N,
            1,
            np.zeros(N),
            [[]] * N,
            hyp_dict,
        )
        K_n_inv = np.linalg.inv(
            K_n + hyp_dict.sigma_2 * np.eye(N)
        )
        v_prod = k_prod - np.matmul(
            np.matmul(k_n_prod.T, K_n_inv), k_n_prod
        )
        return v_prod

    def pred_mean_var(self, domain, X, y_nn, hyp_dict,
                     return_full=False):
        """
        mean and var for test points
        Used in baseline acquisition functions
        
        """
        N, D = np.shape(X)
        M, _ = np.shape(domain)

        k_same = self.covariance_partials(0, 0, np.zeros(D), hyp_dict)

        K_n = self.kmm_extended_gp(X, hyp_dict, m_deg=0)
        k_n_prod = self.k_cross_faster(
            domain, X, hyp_dict.array, hyp_dict.key_dict, hyp_dict
        )
        K_n_inv = np.linalg.inv(
            K_n + hyp_dict.sigma_2 * np.eye(N) \
                + 0.00001 * np.eye(N) # for numerical stability
        )
        shared_term = np.matmul(k_n_prod.T, K_n_inv)
        pred_mean = np.matmul(shared_term, y_nn)
        
        if return_full:
            k_prod = self.kmm_extended_gp(domain, hyp_dict, m_deg=0)
            pred_var = k_prod - np.matmul(shared_term, k_n_prod)
        else:
            pred_var = np.array([
                k_same - np.matmul(shared_term[ii,:], k_n_prod[:,ii])for ii in range(M)])
        return pred_mean, pred_var
    
    def pred_internal_cholesky(self, domain, X, y_nn, hyp_dict):
        N, D = np.shape(X)
        M, _ = np.shape(domain)
        k_prod = self.kmm_extended_gp(domain, hyp_dict, m_deg=0)

        K_n = self.kmm_extended_gp(X, hyp_dict, m_deg=0)
        k_n_prod = self.k_cross_faster(
            domain, X, hyp_dict.array, hyp_dict.key_dict, hyp_dict
        )
        L = np.linalg.cholesky(K_n + hyp_dict.sigma_2 * np.eye(N) \
                + 0.00001 * np.eye(N) # for numerical stability
        )

        alpha = np.linalg.solve(L.T, np.linalg.solve(L, y_nn))
        vv = np.linalg.solve(L, k_n_prod)
        
        pred_mean = np.matmul(k_n_prod.T, alpha)
        
        pred_var = k_prod - np.matmul(vv.T,vv)
        return pred_mean, pred_var
        
        
    def pred_mean_var_cholesky(self, domain, X, y_nn, hyp_dict):
        """
        mean and var for test points
        Used in baseline acquisition functions
        """
        pred_mean, pred_var = self.pred_internal_cholesky(
            domain, X, y_nn, hyp_dict)

        return pred_mean, np.diag(pred_var)
    
    def pred_post_sample(self, domain, X, y_nn, hyp_dict):
        pred_mean, pred_var = self.pred_internal_cholesky(
            domain, X, y_nn, hyp_dict)
        return np.random.multivariate_normal(pred_mean, pred_var)

    def knn_with_noise(self, domain, hyp_dict):
        K_n = self.kmm_extended_gp(domain, hyp_dict)
        N, _ = np.shape(K_n)
        return K_n + hyp_dict.sigma_2 * np.eye(N) \
                   + + 0.00001 * np.eye(N) # for numerical stability

class RBF_kernel(Kernel):
    """Kernel with RBF covariance"""
    def __init__(self):
        self.name = "RBF"

    def alpha(self, hyp_dict):
        """Used for sampling maximum"""
        return hyp_dict.sigma_rbf

    def sigma(self, hyp_dict):
        """Used in Fourier transform"""
        return np.diag(2 / hyp_dict.ll ** 2)

    def cov_p_w(self, hyp_dict, dom_size, D):
        """Sample w for maximum sampling"""
        mean = np.zeros(D)
        Sigma = self.sigma(hyp_dict)
        return np.random.multivariate_normal(
            mean, Sigma, dom_size
        )

    def base_multi(self, domain, hyp_vec, key_dict, hyp_dict):
        """Base covariance for multiple points.
        Used when calculating log probability (can't use
        stored values)"""
        nn, D = np.shape(domain)

        tmp = np.broadcast_to(domain.T, (nn, D, nn))
        tmp = np.transpose(tmp, [0, 2, 1])
        diff = tmp - np.transpose(tmp, [1, 0, 2])
        
        sigma_rbf = self.get_var("sigma_rbf", hyp_vec, key_dict, hyp_dict)
        ll = self.get_var("ll", hyp_vec, key_dict, hyp_dict)
        
        return sigma_rbf * np.exp(
            -np.sum(diff ** 2 / ll ** 2, 2)
        )
    
    def k_cross_faster(self, domain, training, 
                       hyp_vec, key_dict, hyp_dict):
        """Base covariance for multiple points.
        Used when calculating log probability (can't use
        stored values)"""
        mm, D = np.shape(domain)
        nn, _ = np.shape(training)
        
        tmp = np.broadcast_to(domain.T, (nn, D, mm))
        tmp = np.transpose(tmp, [0, 2, 1])
        
        tmp_train = np.broadcast_to(training.T, (mm, D, nn))
        tmp_train = np.transpose(tmp_train, [2, 0, 1])
        
        diff = tmp - tmp_train
        
        sigma_rbf = self.get_var("sigma_rbf", hyp_vec, key_dict, hyp_dict)
        ll = self.get_var("ll", hyp_vec, key_dict, hyp_dict)

        return sigma_rbf * np.exp(
            -np.sum(diff ** 2 / ll ** 2, 2)
        )

    
    def covariance_partials(
        self, x1_pp, x2_pp, tau, hyp_dict, dd_dims=None
    ):
        """
        RBF covaraiance function for different degrees of
        differentiation
        
        x1_pp: scalar, derivative-degree of point 1
        x2_pp: scalar, derivative-degree of point 2
        tau:   vector, distance between points
        hyp_dict: dict of hyperparameters
        dd_1:  scalar, first dimension for derivation
        dd_2:  scalar, second dimenion for derivation
        dd_3:  scalar, third dimension for derivation
        dd_4:  scalar, fourth dimenion for derivation
        
        """
        # tau = x1-x2
        ll = hyp_dict.ll  # vector
        def base(tau, ll):  # tau and ll are numpy arrays
            """Covariance function with no differentiation"""
            return hyp_dict.sigma_rbf * np.exp(
                -np.sum(tau ** 2 / ll ** 2, 0)
            )

        # only some cases are needed hence implemented
        assert x1_pp + x2_pp <= 4
        assert x1_pp >= 0
        assert x2_pp >= 0
        assert x1_pp <= 2
        assert x2_pp <= 2

        def dev_x1(tau, ll, dd_1):
            """x1 differentiated"""
            return -2 / ll[dd_1] ** 2 * tau[dd_1] * base(tau, ll)

        def dev_x1x2(tau, ll, dd_1, dd_2):
            """x1 and x2 differentiated"""
            if dd_1 == dd_2:
                return (
                    -2
                    / ll[dd_1] ** 2
                    * (-1 + 2 / ll[dd_1] ** 2 * tau[dd_1] ** 2)
                    * base(tau, ll)
                )
            else:
                return (
                    -4
                    * tau[dd_1]
                    / ll[dd_1] ** 2
                    * tau[dd_2]
                    / ll[dd_2] ** 2
                    * base(tau, ll)
                )

        def dev_x1_x1_x2(tau, ll, dd_1, dd_2, dd_3):
            """x1 differentiated once, x2 twice"""
            if dd_1 == dd_2 == dd_3:
                return (
                    4
                    / ll[dd_1] ** 4
                    * (
                        -3 * tau[dd_1]
                        + 2 / ll[dd_1] ** 2 * tau[dd_1] ** 3
                    )
                    * base(tau, ll)
                )
            elif dd_1 != dd_2 and dd_1 != dd_3 and dd_2 != dd_3:
                return (
                    6
                    * tau[dd_1]
                    / ll[dd_1] ** 2
                    * tau[dd_2]
                    / ll[dd_2] ** 2
                    * tau[dd_3]
                    / ll[dd_3] ** 2
                    * base(tau, ll)
                )
            else:
                if dd_1 == dd_2:
                    doub = dd_1
                    sing = dd_3
                elif dd_1 == dd_3:
                    doub = dd_1
                    sing = dd_2
                else:
                    doub = dd_2
                    sing = dd_1
                return (
                    -4
                    / ll[doub] ** 2
                    * tau[sing]
                    / ll[sing] ** 2
                    * (1 - 2 * tau[doub] ** 2 / ll[doub] ** 2)
                    * base(tau, ll)
                )

        def dev_x1x1_x2x2(tau, ll, dd_1, dd_2, dd_3, dd_4):
            """x1 and x2 differentiated twice"""
            uniq_dd, count_dd = np.unique(
                np.array([dd_1, dd_2, dd_3, dd_4]),
                return_counts=True,
            )
            max_count = np.max(count_dd)
            if max_count == 1:
                return (
                    2
                    * tau[dd_4]
                    / ll[dd_4] ** 2
                    * dev_x1_x1_x2(tau, ll, dd_1, dd_2, dd_3)
                )
            elif max_count == 3:
                trip = uniq_dd[np.argmax(count_dd)]
                sing = uniq_dd[np.argmin(count_dd)]
                return (
                    2
                    * tau[sing]
                    / ll[sing] ** 2
                    * dev_x1_x1_x2(tau, ll, trip, trip, trip)
                )
            elif max_count == 4:
                return (
                    4
                    / ll[dd_1] ** 4
                    * (
                        3
                        - 12 / ll[dd_1] ** 2 * tau[dd_1] ** 2
                        + 4 / ll[dd_1] ** 4 * tau[dd_1] ** 4
                    )
                    * base(tau, ll)
                )

            if len(uniq_dd) == 2:  # 2+2
                firs = uniq_dd[0]
                seco = uniq_dd[1]

                return (
                    -4
                    / (ll[firs] ** 2 * ll[seco] ** 2)
                    * (
                        -1
                        + 2 / ll[firs] ** 2 * tau[firs] ** 2
                        + 2 / ll[seco] ** 2 * tau[seco] ** 2
                        - 4
                        / (ll[firs] ** 2 * ll[seco] ** 2)
                        * tau[firs] ** 2
                        * tau[seco] ** 2
                    )
                    * base(tau, ll)
                )

            else:  # 2 + 1 + 1
                doub = uniq_dd[np.argmax(count_dd)]
                sing = [ii for ii in uniq_dd if ii != doub]
                return (
                    2
                    * tau[sing[0]]
                    / ll[sing[0]] ** 2
                    * dev_x1_x1_x2(tau, ll, doub, doub, sing[1])
                )

        # Select correct level of differentiation and dimension
        if (x1_pp, x2_pp) == (0, 0):
            return base(tau, ll)
        elif (x1_pp, x2_pp) == (1, 0):
            return dev_x1(tau, ll, dd_dims[0])
        elif (x1_pp, x2_pp) == (0, 1):
            return -dev_x1(tau, ll, dd_dims[0])
        elif (x1_pp, x2_pp) == (2, 0):
            return -dev_x1x2(tau, ll, dd_dims[0], dd_dims[1])
        elif (x1_pp, x2_pp) == (0, 2):
            return -dev_x1x2(tau, ll, dd_dims[0], dd_dims[1])
        elif (x1_pp, x2_pp) == (1, 1):
            return dev_x1x2(tau, ll, dd_dims[0], dd_dims[1])
        elif (x1_pp, x2_pp) == (2, 1):
            return dev_x1_x1_x2(
                tau, ll, dd_dims[0], dd_dims[1], dd_dims[2]
            )
        elif (x1_pp, x2_pp) == (1, 2):
            return -dev_x1_x1_x2(
                tau, ll, dd_dims[0], dd_dims[1], dd_dims[2]
            )
        elif (x1_pp, x2_pp) == (2, 2):
            return dev_x1x1_x2x2(
                tau,
                ll,
                dd_dims[0],
                dd_dims[1],
                dd_dims[2],
                dd_dims[3],
            )
        else:
            print(x1_pp, x2_pp)
            raise ValueError


class Directed_kernel(Kernel):
    """Kernel with directed covariance"""

    def __init__(self):
        self.name = "Directed"

    def alpha(self, hyp_dict):
        """Used for sampling maximum"""
        return hyp_dict.sigma_dir

    def compute_A(self, hyp_vec, key_dict):
        """Compute A on the fly, used by base_multi"""
        gamma = hyp_vec[key_dict["direc"]]
        b_0 = np.cos(gamma)
        b_1 = np.sin(gamma)
        B_0 = 1 - b_0 ** 2
        B_1 = 1 - b_1 ** 2
        B_2 = -b_0 * b_1
        return np.array([[B_0, B_2], [B_2, B_1]])

    def sigma(self, hyp_dict):
        """Used in Fourier transform"""
        return 2 / hyp_dict.ll_dir ** 2 * hyp_dict.A

    def cov_p_w(self, hyp_dict, dom_size, D):
        """Sample w for maximum sampling"""
        mean = np.zeros(D)
        Sigma = self.sigma(hyp_dict)
        return np.random.multivariate_normal(
            mean, Sigma, dom_size
        )

    def base(self, p0, p1, hyp_dict):
        """Covariance function with no differentiation"""
        tau = p1 - p0
        dir_dist = np.matmul(np.matmul(tau.T, hyp_dict.A), tau)
        return hyp_dict.sigma_dir * np.exp(
            -np.sum(dir_dist / hyp_dict.ll_dir ** 2)
        )

    def base_multi(self, domain, hyp_vec, key_dict, hyp_dict):
        """Base covariance for multiple points.
        Used when calculating log probability (can't use
        stored values)"""
        nn, _ = np.shape(domain)

        tmp = np.broadcast_to(domain.T, (nn, 2, nn))
        tmp = np.transpose(tmp, [0, 2, 1])
        diff = tmp - np.transpose(tmp, [1, 0, 2])
        diff_v = np.reshape(diff, (nn ** 2, 2))
        diff_dir = (
            np.matmul(diff_v, self.compute_A(hyp_vec, key_dict))
            * diff_v
        ).sum(1)
        tau_tilde = np.reshape(diff_dir, (nn, nn))
        
        sigma_dir = self.get_var("sigma_dir", hyp_vec, key_dict, hyp_dict)
        ll_dir = self.get_var("ll_dir", hyp_vec, key_dict, hyp_dict)
        
        return sigma_dir * np.exp(
            -tau_tilde / ll_dir ** 2
        )

    def k_cross_faster(self, domain, training, 
                       hyp_vec, key_dict, hyp_dict):
        """Base covariance for multiple points.
        Used when calculating log probability (can't use
        stored values)"""
        mm, _ = np.shape(domain)
        nn, _ = np.shape(training)
        
        tmp = np.broadcast_to(domain.T, (nn, 2, mm))
        tmp = np.transpose(tmp, [0, 2, 1])
        
        tmp_train = np.broadcast_to(training.T, (mm, 2, nn))
        tmp_train = np.transpose(tmp_train, [2, 0, 1])
        
        diff = tmp - tmp_train

        diff_v = np.reshape(diff, (mm * nn, 2))
        diff_dir = (
            np.matmul(diff_v, self.compute_A(hyp_vec, key_dict))
            * diff_v
        ).sum(1)
        tau_tilde = np.reshape(diff_dir, (nn, mm))
        
        sigma_dir = self.get_var("sigma_dir", hyp_vec, key_dict, hyp_dict)
        ll_dir = self.get_var("ll_dir", hyp_vec, key_dict, hyp_dict)

        return sigma_dir * np.exp(
            -tau_tilde / ll_dir ** 2
        )
    
    def covariance_partials(
        self, x1_pp, x2_pp, tau, hyp_dict, dd_dims=None
    ):
        """
        Directed covaraiance function for different degrees of
        differentiation
        x1_pp: scalar, derivative-degree of point 1
        x2_pp: scalar, derivative-degree of point 2
        tau:   vector, distance between points
        hyp_dict: dict of hyperparameters
        dd_1:  scalar, first dimension for derivation
        dd_2:  scalar, second dimenion for derivation
        dd_3:  scalar, third dimension for derivation
        dd_4:  scalar, fourth dimenion for derivation
        
        """


        # only some cases are implemented
        assert x1_pp + x2_pp <= 4
        assert x1_pp >= 0
        assert x2_pp >= 0
        assert x1_pp <= 2
        assert x2_pp <= 2

        def term_0_fn(tau, hyp_dict):
            """Term used multiple times relating to dim 0"""
            return tau[0] * hyp_dict.B_0 + tau[1] * hyp_dict.B_2

        def term_1_fn(tau, hyp_dict):
            """Term used multiple times relating to dim 1"""
            return tau[0] * hyp_dict.B_2 + tau[1] * hyp_dict.B_1

        def deriv_t0(tau, hyp_dict, term_0):
            """Dim 0 differentiated once"""
            return (
                -2
                / hyp_dict.ll_dir ** 2
                * term_0
                * self.base(tau, 0, hyp_dict)
            )

        def deriv_t1(tau, hyp_dict, term_1):
            """Dim 1 differentiated once"""
            return (
                -2
                / hyp_dict.ll_dir ** 2
                * term_1
                * self.base(tau, 0, hyp_dict)
            )

        def deriv_t0t0(tau, hyp_dict, term_0):
            """Dim 0 differentiated twice"""
            return (
                2
                / hyp_dict.ll_dir ** 2
                * (
                    2 / hyp_dict.ll_dir ** 2 * term_0 ** 2
                    - hyp_dict.B_0
                )
                * self.base(tau, 0, hyp_dict)
            )

        def deriv_t1t1(tau, hyp_dict, term_1):
            """Dim 1 differentiated twice"""
            return (
                2
                / hyp_dict.ll_dir ** 2
                * (
                    2 / hyp_dict.ll_dir ** 2 * term_1 ** 2
                    - hyp_dict.B_1
                )
                * self.base(tau, 0, hyp_dict)
            )

        def deriv_t0t1(tau, hyp_dict, term_0, term_1):
            """Dim 0 and 1 differentiated once"""
            return (
                2
                / hyp_dict.ll_dir ** 2
                * (
                    2 / hyp_dict.ll_dir ** 2 * term_1 * term_0
                    - hyp_dict.B_2
                )
                * self.base(tau, 0, hyp_dict)
            )

        def deriv_t0t0t0(tau, hyp_dict, term_0):
            """Dim 0 differentiated thrice"""
            return (
                4
                / hyp_dict.ll_dir ** 4
                * (
                    -2 / hyp_dict.ll_dir ** 2 * term_0 ** 3
                    + 3 * hyp_dict.B_0 * term_0
                )
                * self.base(tau, 0, hyp_dict)
            )

        def deriv_t1t1t1(tau, hyp_dict, term_1):
            """Dim 1 differentiated thrice"""
            return (
                4
                / hyp_dict.ll_dir ** 4
                * (
                    -2 / hyp_dict.ll_dir ** 2 * term_1 ** 3
                    + 3 * hyp_dict.B_1 * term_1
                )
                * self.base(tau, 0, hyp_dict)
            )

        def deriv_t0t0t1(tau, hyp_dict, term_0, term_1):
            """Dim 0 differentiated twice and dim 1
            differentiated once"""
            return (
                4
                / hyp_dict.ll_dir ** 4
                * (
                    -2
                    / hyp_dict.ll_dir ** 2
                    * term_0 ** 2
                    * term_1
                    + hyp_dict.B_0 * term_1
                    + 2 * hyp_dict.B_2 * term_0
                )
                * self.base(tau, 0, hyp_dict)
            )

        def deriv_t0t1t1(tau, hyp_dict, term_0, term_1):
            """Dim 0 differentiated once and dim 1
            differentiated twice"""
            return (
                4
                / hyp_dict.ll_dir ** 4
                * (
                    -2
                    / hyp_dict.ll_dir ** 2
                    * term_1 ** 2
                    * term_0
                    + hyp_dict.B_1 * term_0
                    + 2 * hyp_dict.B_2 * term_1
                )
                * self.base(tau, 0, hyp_dict)
            )

        def deriv_t0t0t0t0(tau, hyp_dict, term_0):
            """Dim 0 differentiated four times"""
            return (
                4
                / hyp_dict.ll_dir ** 4
                * (
                    4 / hyp_dict.ll_dir ** 4 * term_0 ** 4
                    - 12
                    / hyp_dict.ll_dir ** 2
                    * hyp_dict.B_0
                    * term_0 ** 2
                    + 3 * hyp_dict.B_0 ** 2
                )
                * self.base(tau, 0, hyp_dict)
            )

        def deriv_t1t1t1t1(tau, hyp_dict, term_1):
            """Dim 1 differentiated four times"""
            return (
                4
                / hyp_dict.ll_dir ** 4
                * (
                    4 / hyp_dict.ll_dir ** 4 * term_1 ** 4
                    - 12
                    / hyp_dict.ll_dir ** 2
                    * hyp_dict.B_1
                    * term_1 ** 2
                    + 3 * hyp_dict.B_1 ** 2
                )
                * self.base(tau, 0, hyp_dict)
            )

        def deriv_t0t0t0t1(tau, hyp_dict, term_0, term_1):
            """Dim 0 differentiated thrice and dim 1
            differentiated once"""
            return (
                4
                / hyp_dict.ll_dir ** 4
                * (
                    4
                    / hyp_dict.ll_dir ** 4
                    * term_0 ** 3
                    * term_1
                    - 6
                    / hyp_dict.ll_dir ** 2
                    * term_0
                    * (
                        hyp_dict.B_0 * term_1
                        + hyp_dict.B_2 * term_0
                    )
                    + 3 * hyp_dict.B_0 * hyp_dict.B_2
                )
                * self.base(tau, 0, hyp_dict)
            )

        def deriv_t0t1t1t1(tau, hyp_dict, term_0, term_1):
            """Dim 0 differentiated once and dim 1
            differentiated thrice"""
            return (
                4
                / hyp_dict.ll_dir ** 4
                * (
                    4
                    / hyp_dict.ll_dir ** 4
                    * term_1 ** 3
                    * term_0
                    - 6
                    / hyp_dict.ll_dir ** 2
                    * term_1
                    * (
                        hyp_dict.B_1 * term_0
                        + hyp_dict.B_2 * term_1
                    )
                    + 3 * hyp_dict.B_1 * hyp_dict.B_2
                )
                * self.base(tau, 0, hyp_dict)
            )

        def deriv_t0t0t1t1(tau, hyp_dict, term_0, term_1):
            """Dim 0 differentiated twice and dim 1
            differentiated twice"""
            return (
                4
                / hyp_dict.ll_dir ** 4
                * (
                    4
                    / hyp_dict.ll_dir ** 4
                    * term_1 ** 2
                    * term_0 ** 2
                    - 2
                    / hyp_dict.ll_dir ** 2
                    * hyp_dict.B_1
                    * term_0 ** 2
                    - 2
                    / hyp_dict.ll_dir ** 2
                    * hyp_dict.B_0
                    * term_1 ** 2
                    - 8
                    / hyp_dict.ll_dir ** 2
                    * hyp_dict.B_2
                    * term_0
                    * term_1
                    + hyp_dict.B_0 * hyp_dict.B_1
                    + 2 * hyp_dict.B_2 ** 2
                )
                * self.base(tau, 0, hyp_dict)
            )

        D = len(tau)

        # Pick the correct level of differentiation and dimensions
        if dd_dims is None or len(dd_dims) == 0:
            return self.base(tau, np.zeros(D), hyp_dict)
        else:
            sign = -2 * (x2_pp % 2 == 1) + 1
            if len(dd_dims) == 1:
                if dd_dims[0] == 0:
                    term_0 = term_0_fn(tau, hyp_dict)
                    return sign * deriv_t0(tau, hyp_dict, term_0)
                else:
                    term_1 = term_1_fn(tau, hyp_dict)
                    return sign * deriv_t1(tau, hyp_dict, term_1)
            elif len(dd_dims) == 2:
                if np.sum(dd_dims) == 0:
                    term_0 = term_0_fn(tau, hyp_dict)
                    return sign * deriv_t0t0(
                        tau, hyp_dict, term_0
                    )
                elif np.sum(dd_dims) == 1:
                    term_0 = term_0_fn(tau, hyp_dict)
                    term_1 = term_1_fn(tau, hyp_dict)
                    return sign * deriv_t0t1(
                        tau, hyp_dict, term_0, term_1
                    )
                else:
                    term_1 = term_1_fn(tau, hyp_dict)
                    return sign * deriv_t1t1(
                        tau, hyp_dict, term_1
                    )
            elif len(dd_dims) == 3:
                if np.sum(dd_dims) == 0:
                    term_0 = term_0_fn(tau, hyp_dict)
                    return sign * deriv_t0t0t0(
                        tau, hyp_dict, term_0
                    )
                elif np.sum(dd_dims) == 1:
                    term_0 = term_0_fn(tau, hyp_dict)
                    term_1 = term_1_fn(tau, hyp_dict)
                    return sign * deriv_t0t0t1(
                        tau, hyp_dict, term_0, term_1
                    )
                elif np.sum(dd_dims) == 2:
                    term_0 = term_0_fn(tau, hyp_dict)
                    term_1 = term_1_fn(tau, hyp_dict)
                    return sign * deriv_t0t1t1(
                        tau, hyp_dict, term_0, term_1
                    )
                else:
                    term_1 = term_1_fn(tau, hyp_dict)
                    return sign * deriv_t1t1t1(
                        tau, hyp_dict, term_1
                    )
            else:
                if np.sum(dd_dims) == 0:
                    term_0 = term_0_fn(tau, hyp_dict)
                    return sign * deriv_t0t0t0t0(
                        tau, hyp_dict, term_0
                    )
                elif np.sum(dd_dims) == 1:
                    term_0 = term_0_fn(tau, hyp_dict)
                    term_1 = term_1_fn(tau, hyp_dict)
                    return sign * deriv_t0t0t0t1(
                        tau, hyp_dict, term_0, term_1
                    )
                elif np.sum(dd_dims) == 2:
                    term_0 = term_0_fn(tau, hyp_dict)
                    term_1 = term_1_fn(tau, hyp_dict)
                    return sign * deriv_t0t0t1t1(
                        tau, hyp_dict, term_0, term_1
                    )
                elif np.sum(dd_dims) == 3:
                    term_0 = term_0_fn(tau, hyp_dict)
                    term_1 = term_1_fn(tau, hyp_dict)
                    return sign * deriv_t0t1t1t1(
                        tau, hyp_dict, term_0, term_1
                    )
                else:
                    term_1 = term_1_fn(tau, hyp_dict)
                    return sign * deriv_t1t1t1t1(
                        tau, hyp_dict, term_1
                    )
                

class Wind_product_kernel(Kernel):
    """Kernel that is product between directed and RBF kernels"""

    def __init__(self):
        self.name = "Product"

    def alpha(self, hyp_dict):
        """Used for sampling maximum"""
        return hyp_dict.sigma_hat

    def compute_A(self, hyp_vec, key_dict):
        """Compute A on the fly, used by base_multi"""
        gamma = hyp_vec[key_dict["direc"]]
        b_0 = np.cos(gamma)
        b_1 = np.sin(gamma)
        B_0 = 1 - b_0 ** 2
        B_1 = 1 - b_1 ** 2
        B_2 = -b_0 * b_1
        return np.array([[B_0, B_2], [B_2, B_1]])
    
    def compute_E(self, hyp_vec, key_dict):
        """Compute E on the fly, used by base_multi"""
        A = self.compute_A(hyp_vec, key_dict)
        ll_ratio = hyp_vec[key_dict['ll_dir']]**2 / hyp_vec[key_dict['ll']]**2
        E_0 = A[0,0] + ll_ratio
        E_1 = A[1,1] + ll_ratio
        return np.array([[E_0, A[0,1]], [A[1,0], E_1]])

    def sigma(self, hyp_dict):
        """Used in Fourier transform"""
        return 2 / hyp_dict.ll_dir ** 2 * hyp_dict.E

    def cov_p_w(self, hyp_dict, dom_size, D):
        """Sample w for maximum sampling"""
        mean = np.zeros(D)
        Sigma = self.sigma(hyp_dict)
        return np.random.multivariate_normal(
            mean, Sigma, dom_size
        )

    def base(self, p0, p1, hyp_dict):
        """Covariance function with no differentiation"""
        tau = p1 - p0
        dir_dist = np.matmul(np.matmul(tau.T, hyp_dict.E), tau)
        return hyp_dict.sigma_hat * np.exp(
            -np.sum(dir_dist / hyp_dict.ll_dir ** 2)
        )

    def base_multi(self, domain, hyp_vec, key_dict, hyp_dict):
        """Base covariance for multiple points.
        Used when calculating log probability (can't use
        stored values)"""
        nn, _ = np.shape(domain)

        tmp = np.broadcast_to(domain.T, (nn, 2, nn))
        tmp = np.transpose(tmp, [0, 2, 1])
        diff = tmp - np.transpose(tmp, [1, 0, 2])
        diff_v = np.reshape(diff, (nn ** 2, 2))
        diff_dir = (
            np.matmul(diff_v, self.compute_E(hyp_vec, key_dict))
            * diff_v
        ).sum(1)
        tau_tilde = np.reshape(diff_dir, (nn, nn))
        
        sigma_hat = self.get_var("sigma_hat", hyp_vec, key_dict, hyp_dict)
        ll_dir = self.get_var("ll_dir", hyp_vec, key_dict, hyp_dict)

        return sigma_hat * np.exp(
            -tau_tilde / ll_dir ** 2
        )

    def k_cross_faster(self, domain, training, 
                       hyp_vec, key_dict, hyp_dict):
        """Base covariance for multiple points.
        Used when calculating log probability (can't use
        stored values)"""
        mm, _ = np.shape(domain)
        nn, _ = np.shape(training)
        
        tmp = np.broadcast_to(domain.T, (nn, 2, mm))
        tmp = np.transpose(tmp, [0, 2, 1])
        
        tmp_train = np.broadcast_to(training.T, (mm, 2, nn))
        tmp_train = np.transpose(tmp_train, [2, 0, 1])
        
        diff = tmp - tmp_train
        diff_v = np.reshape(diff, (nn * mm, 2))
        diff_dir = (
            np.matmul(diff_v, self.compute_E(hyp_vec, key_dict))
            * diff_v
        ).sum(1)
        tau_tilde = np.reshape(diff_dir, (nn, mm))
        sigma_hat = self.get_var("sigma_hat", hyp_vec, key_dict, hyp_dict)
        ll_dir = self.get_var("ll_dir", hyp_vec, key_dict, hyp_dict)

        return sigma_hat * np.exp(
            -tau_tilde / ll_dir ** 2
        )
    
    
    def covariance_partials(
        self, x1_pp, x2_pp, tau, hyp_dict, dd_dims=None
    ):
        """
        Directed covaraiance function for different degrees of
        differentiation
        x1_pp: scalar, derivative-degree of point 1
        x2_pp: scalar, derivative-degree of point 2
        tau:   vector, distance between points
        hyp_dict: dict of hyperparameters
        dd_1:  scalar, first dimension for derivation
        dd_2:  scalar, second dimenion for derivation
        dd_3:  scalar, third dimension for derivation
        dd_4:  scalar, fourth dimenion for derivation
        
        """

        # only some cases are implemented
        assert x1_pp + x2_pp <= 4
        assert x1_pp >= 0
        assert x2_pp >= 0
        assert x1_pp <= 2
        assert x2_pp <= 2

        def term_0_fn(tau, hyp_dict):
            """Term used multiple times relating to dim 0"""
            return tau[0] * hyp_dict.E_0 + tau[1] * hyp_dict.E_2

        def term_1_fn(tau, hyp_dict):
            """Term used multiple times relating to dim 1"""
            return tau[0] * hyp_dict.E_2 + tau[1] * hyp_dict.E_1

        def deriv_t0(tau, hyp_dict, term_0):
            """Dim 0 differentiated once"""
            return (
                -2
                / hyp_dict.ll_dir ** 2
                * term_0
                * self.base(tau, 0, hyp_dict)
            )

        def deriv_t1(tau, hyp_dict, term_1):
            """Dim 1 differentiated once"""
            return (
                -2
                / hyp_dict.ll_dir ** 2
                * term_1
                * self.base(tau, 0, hyp_dict)
            )

        def deriv_t0t0(tau, hyp_dict, term_0):
            """Dim 0 differentiated twice"""
            return (
                2
                / hyp_dict.ll_dir ** 2
                * (
                    2 / hyp_dict.ll_dir ** 2 * term_0 ** 2
                    - hyp_dict.E_0
                )
                * self.base(tau, 0, hyp_dict)
            )

        def deriv_t1t1(tau, hyp_dict, term_1):
            """Dim 1 differentiated twice"""
            return (
                2
                / hyp_dict.ll_dir ** 2
                * (
                    2 / hyp_dict.ll_dir ** 2 * term_1 ** 2
                    - hyp_dict.E_1
                )
                * self.base(tau, 0, hyp_dict)
            )

        def deriv_t0t1(tau, hyp_dict, term_0, term_1):
            """Dim 0 and 1 differentiated once"""
            return (
                2
                / hyp_dict.ll_dir ** 2
                * (
                    2 / hyp_dict.ll_dir ** 2 * term_1 * term_0
                    - hyp_dict.E_2
                )
                * self.base(tau, 0, hyp_dict)
            )

        def deriv_t0t0t0(tau, hyp_dict, term_0):
            """Dim 0 differentiated thrice"""
            return (
                4
                / hyp_dict.ll_dir ** 4
                * (
                    -2 / hyp_dict.ll_dir ** 2 * term_0 ** 3
                    + 3 * hyp_dict.E_0 * term_0
                )
                * self.base(tau, 0, hyp_dict)
            )

        def deriv_t1t1t1(tau, hyp_dict, term_1):
            """Dim 1 differentiated thrice"""
            return (
                4
                / hyp_dict.ll_dir ** 4
                * (
                    -2 / hyp_dict.ll_dir ** 2 * term_1 ** 3
                    + 3 * hyp_dict.E_1 * term_1
                )
                * self.base(tau, 0, hyp_dict)
            )

        def deriv_t0t0t1(tau, hyp_dict, term_0, term_1):
            """Dim 0 differentiated twice and dim 1
            differentiated once"""
            return (
                4
                / hyp_dict.ll_dir ** 4
                * (
                    -2
                    / hyp_dict.ll_dir ** 2
                    * term_0 ** 2
                    * term_1
                    + hyp_dict.E_0 * term_1
                    + 2 * hyp_dict.E_2 * term_0
                )
                * self.base(tau, 0, hyp_dict)
            )

        def deriv_t0t1t1(tau, hyp_dict, term_0, term_1):
            """Dim 0 differentiated once and dim 1
            differentiated twice"""
            return (
                4
                / hyp_dict.ll_dir ** 4
                * (
                    -2
                    / hyp_dict.ll_dir ** 2
                    * term_1 ** 2
                    * term_0
                    + hyp_dict.E_1 * term_0
                    + 2 * hyp_dict.E_2 * term_1
                )
                * self.base(tau, 0, hyp_dict)
            )

        def deriv_t0t0t0t0(tau, hyp_dict, term_0):
            """Dim 0 differentiated four times"""
            return (
                4
                / hyp_dict.ll_dir ** 4
                * (
                    4 / hyp_dict.ll_dir ** 4 * term_0 ** 4
                    - 12
                    / hyp_dict.ll_dir ** 2
                    * hyp_dict.E_0
                    * term_0 ** 2
                    + 3 * hyp_dict.E_0 ** 2
                )
                * self.base(tau, 0, hyp_dict)
            )

        def deriv_t1t1t1t1(tau, hyp_dict, term_1):
            """Dim 1 differentiated four times"""
            return (
                4
                / hyp_dict.ll_dir ** 4
                * (
                    4 / hyp_dict.ll_dir ** 4 * term_1 ** 4
                    - 12
                    / hyp_dict.ll_dir ** 2
                    * hyp_dict.E_1
                    * term_1 ** 2
                    + 3 * hyp_dict.E_1 ** 2
                )
                * self.base(tau, 0, hyp_dict)
            )

        def deriv_t0t0t0t1(tau, hyp_dict, term_0, term_1):
            """Dim 0 differentiated thrice and dim 1
            differentiated once"""
            return (
                4
                / hyp_dict.ll_dir ** 4
                * (
                    4
                    / hyp_dict.ll_dir ** 4
                    * term_0 ** 3
                    * term_1
                    - 6
                    / hyp_dict.ll_dir ** 2
                    * term_0
                    * (
                        hyp_dict.E_0 * term_1
                        + hyp_dict.E_2 * term_0
                    )
                    + 3 * hyp_dict.E_0 * hyp_dict.E_2
                )
                * self.base(tau, 0, hyp_dict)
            )

        def deriv_t0t1t1t1(tau, hyp_dict, term_0, term_1):
            """Dim 0 differentiated once and dim 1
            differentiated thrice"""
            return (
                4
                / hyp_dict.ll_dir ** 4
                * (
                    4
                    / hyp_dict.ll_dir ** 4
                    * term_1 ** 3
                    * term_0
                    - 6
                    / hyp_dict.ll_dir ** 2
                    * term_1
                    * (
                        hyp_dict.E_1 * term_0
                        + hyp_dict.E_2 * term_1
                    )
                    + 3 * hyp_dict.E_1 * hyp_dict.E_2
                )
                * self.base(tau, 0, hyp_dict)
            )

        def deriv_t0t0t1t1(tau, hyp_dict, term_0, term_1):
            """Dim 0 differentiated twice and dim 1
            differentiated twice"""
            return (
                4
                / hyp_dict.ll_dir ** 4
                * (
                    4
                    / hyp_dict.ll_dir ** 4
                    * term_1 ** 2
                    * term_0 ** 2
                    - 2
                    / hyp_dict.ll_dir ** 2
                    * hyp_dict.E_1
                    * term_0 ** 2
                    - 2
                    / hyp_dict.ll_dir ** 2
                    * hyp_dict.E_0
                    * term_1 ** 2
                    - 8
                    / hyp_dict.ll_dir ** 2
                    * hyp_dict.E_2
                    * term_0
                    * term_1
                    + hyp_dict.E_0 * hyp_dict.E_1
                    + 2 * hyp_dict.E_2 ** 2
                )
                * self.base(tau, 0, hyp_dict)
            )

        D = len(tau)

        # Pick the correct level of differentiation and dimensions
        if dd_dims is None or len(dd_dims) == 0:
            return self.base(tau, np.zeros(D), hyp_dict)
        else:
            sign = -2 * (x2_pp % 2 == 1) + 1
            if len(dd_dims) == 1:
                if dd_dims[0] == 0:
                    term_0 = term_0_fn(tau, hyp_dict)
                    return sign * deriv_t0(tau, hyp_dict, term_0)
                else:
                    term_1 = term_1_fn(tau, hyp_dict)
                    return sign * deriv_t1(tau, hyp_dict, term_1)
            elif len(dd_dims) == 2:
                if np.sum(dd_dims) == 0:
                    term_0 = term_0_fn(tau, hyp_dict)
                    return sign * deriv_t0t0(
                        tau, hyp_dict, term_0
                    )
                elif np.sum(dd_dims) == 1:
                    term_0 = term_0_fn(tau, hyp_dict)
                    term_1 = term_1_fn(tau, hyp_dict)
                    return sign * deriv_t0t1(
                        tau, hyp_dict, term_0, term_1
                    )
                else:
                    term_1 = term_1_fn(tau, hyp_dict)
                    return sign * deriv_t1t1(
                        tau, hyp_dict, term_1
                    )
            elif len(dd_dims) == 3:
                if np.sum(dd_dims) == 0:
                    term_0 = term_0_fn(tau, hyp_dict)
                    return sign * deriv_t0t0t0(
                        tau, hyp_dict, term_0
                    )
                elif np.sum(dd_dims) == 1:
                    term_0 = term_0_fn(tau, hyp_dict)
                    term_1 = term_1_fn(tau, hyp_dict)
                    return sign * deriv_t0t0t1(
                        tau, hyp_dict, term_0, term_1
                    )
                elif np.sum(dd_dims) == 2:
                    term_0 = term_0_fn(tau, hyp_dict)
                    term_1 = term_1_fn(tau, hyp_dict)
                    return sign * deriv_t0t1t1(
                        tau, hyp_dict, term_0, term_1
                    )
                else:
                    term_1 = term_1_fn(tau, hyp_dict)
                    return sign * deriv_t1t1t1(
                        tau, hyp_dict, term_1
                    )
            else:
                if np.sum(dd_dims) == 0:
                    term_0 = term_0_fn(tau, hyp_dict)
                    return sign * deriv_t0t0t0t0(
                        tau, hyp_dict, term_0
                    )
                elif np.sum(dd_dims) == 1:
                    term_0 = term_0_fn(tau, hyp_dict)
                    term_1 = term_1_fn(tau, hyp_dict)
                    return sign * deriv_t0t0t0t1(
                        tau, hyp_dict, term_0, term_1
                    )
                elif np.sum(dd_dims) == 2:
                    term_0 = term_0_fn(tau, hyp_dict)
                    term_1 = term_1_fn(tau, hyp_dict)
                    return sign * deriv_t0t0t1t1(
                        tau, hyp_dict, term_0, term_1
                    )
                elif np.sum(dd_dims) == 3:
                    term_0 = term_0_fn(tau, hyp_dict)
                    term_1 = term_1_fn(tau, hyp_dict)
                    return sign * deriv_t0t1t1t1(
                        tau, hyp_dict, term_0, term_1
                    )
                else:
                    term_1 = term_1_fn(tau, hyp_dict)
                    return sign * deriv_t1t1t1t1(
                        tau, hyp_dict, term_1
                    )


class Wind_sum_kernel(Kernel):
    """Wind sum kernel: sum of an RBF kernel and a Directed kernel"""


    def __init__(self):
        self.name = "Sum"
        self.kernel_rbf = RBF_kernel()
        self.kernel_dir = Directed_kernel()

    def alpha(self, hyp_dict):
        """Used for sampling maximum"""
        return self.kernel_rbf.alpha(
            hyp_dict
        ) + self.kernel_dir.alpha(hyp_dict)

    def cov_p_w(self, hyp_dict, dom_size, D):
        """Sample w for maximum sampling"""
        mean = np.zeros(D)

        Sigma_rbf = self.kernel_rbf.sigma(hyp_dict)
        Sigma_dir = self.kernel_dir.sigma(hyp_dict)

        return np.random.multivariate_normal(
            mean, Sigma_dir + Sigma_rbf, dom_size
        )

    def base_multi(self, domain, hyp_vec, key_dict, hyp_dict):
        """Base covariance for multiple points.
        Used when calculating log probability (can't use
        stored values)
        Sum of parts"""
        return self.kernel_rbf.base_multi(
            domain, hyp_vec, key_dict, hyp_dict
        ) + self.kernel_dir.base_multi(
            domain, hyp_vec, key_dict, hyp_dict)

    def k_cross_faster(self, domain, training, 
                       hyp_vec, key_dict, hyp_dict):
        return self.kernel_rbf.k_cross_faster(
        domain, training, hyp_vec, key_dict, hyp_dict) \
        + self.kernel_dir.k_cross_faster(
        domain, training, hyp_vec, key_dict, hyp_dict)
    
    
    def covariance_partials(
        self, x1_pp, x2_pp, tau, hyp_dict, dd_dims=None
    ):
        """
        x1_pp: scalar, derivative-degree of point 1
        x2_pp: scalar, derivative-degree of point 2
        tau:   vector, distance between points
        hyp_dict: dict of hyperparameters
        dd_1:  scalar, first dimension for derivation
        dd_2:  scalar, second dimenion for derivation
        dd_3:  scalar, third dimension for derivation
        dd_4:  scalar, fourth dimenion for derivation
        
        Sum of kernel parts
        """
        return self.kernel_rbf.covariance_partials(
            x1_pp, x2_pp, tau, hyp_dict, dd_dims
        ) + self.kernel_dir.covariance_partials(
            x1_pp, x2_pp, tau, hyp_dict, dd_dims
        )
    
class Combination_kernel(Kernel):
    """Combination kernel: sum/product of several kernels"""


    def __init__(self, parts, way='Summation'):
        assert way in ['Summation', 'Multiplication']
        self.name = "Combination"
        self.way = way
        self.kernel_list = []
        for pp in parts:
            if type(pp) is str:
                self.kernel_list.append(choose_kernel(pp))
            else:
                self.kernel_list.append(pp)
    def first_kern(self):
        return self.kernel_list[0]
    def rem_kernels(self):
        return self.kernel_list[1:]

    def alpha(self, hyp_dict):
        """Used for sampling maximum"""
        assert self.way == 'Summation'
        tally = self.first_kern().alpha(hyp_dict.values[0])
        rem_kernels = self.rem_kernels()
        for kk in range(len(rem_kernels)):
            kern = rem_kernels[kk]
            tally += kern.alpha(hyp_dict.values[kk+1])
        return tally
    
    def cov_p_w(self, hyp_dict, dom_size, D):
        """Sample w for maximum sampling"""
        assert self.way == 'Summation'
        mean = np.zeros(D)
        
        Sigma_tally = self.first_kern().sigma(hyp_dict.values[0])
        rem_kernels = self.rem_kernels()
        for kk in range(len(rem_kernels)):
            kern = rem_kernels[kk]
            Sigma_tally += kern.sigma(hyp_dict.values[kk+1])

        return np.random.multivariate_normal(
            mean, Sigma_tally, dom_size
        )

    def base_multi(self, domain, hyp_vec, key_dict, hyp_dict):
        """Base covariance for multiple points.
        Used when calculating log probability (can't use
        stored values)
        Sum of parts"""
        kernel_tally = self.first_kern().base_multi(
            domain, hyp_vec, key_dict[0], hyp_dict.values[0])
        rem_kernels = self.rem_kernels()
        for kk in range(len(rem_kernels)):
            kern = rem_kernels[kk]
            kern_val = kern.base_multi(
                domain, hyp_vec, key_dict[kk+1], hyp_dict.values[kk+1])
            if self.way == 'Summation':
                kernel_tally += kern_val
            elif self.way == 'Multiplication':
                kernel_tally *= kern_val
        return kernel_tally

    def k_cross_faster(self, domain, training, 
                       hyp_vec, key_dict, hyp_dict):
        k_tally = self.first_kern().k_cross_faster(
            domain, training, hyp_vec, key_dict[0], hyp_dict.values[0])
        rem_kernels = self.rem_kernels()
        for kk in range(len(rem_kernels)):
            kern = rem_kernels[kk]
            kern_val = kern.k_cross_faster(
                domain, training, hyp_vec, key_dict[kk+1], hyp_dict.values[kk+1])
            if self.way == 'Summation':
                k_tally += kern_val
            elif self.way == 'Multiplication':
                k_tally *= kern_val
        return k_tally
    
    def covariance_partials(
        self, x1_pp, x2_pp, tau, hyp_dict, dd_dims=None
    ):
        """
        x1_pp: scalar, derivative-degree of point 1
        x2_pp: scalar, derivative-degree of point 2
        tau:   vector, distance between points
        hyp_dict: dict of hyperparameters
        dd_1:  scalar, first dimension for derivation
        dd_2:  scalar, second dimenion for derivation
        dd_3:  scalar, third dimension for derivation
        dd_4:  scalar, fourth dimenion for derivation
        
        Sum of kernel parts
        """
        if self.way == 'Multiplication':
            assert x1_pp == 0
            assert x2_pp == 0
            
        k_tally = self.first_kern().covariance_partials(
            x1_pp, x2_pp, tau, 
            hyp_dict.values[0], dd_dims)
        rem_kernels = self.rem_kernels()
        for kk in range(len(rem_kernels)):
            kern = rem_kernels[kk]
            kern_val = kern.covariance_partials(
                x1_pp, x2_pp, tau, 
                hyp_dict.values[kk+1], dd_dims)
            if self.way == 'Summation':
                k_tally += kern_val
            elif self.way == 'Multiplication':
                k_tally *= kern_val
        return k_tally
        
    
    
class Temporal_kernel(Kernel):
    """Temporal kernel: kernel with one temporal part 
    and one spatial part
    
    Temproal feature has to be the first, 
    and the temporal hyp_dict has to be first.
    """
    
    def __init__(self, temp_kern, spat_kern, way='Multiplication'):
        self.name = "Temporal"
        self.temp_kern = temp_kern
        self.spat_kern = spat_kern
        self.way = way
        assert way in ['Summation', 'Multiplication']

    def alpha(self, hyp_dict):
        """Used for sampling maximum"""
        raise NotImplementedError
    
    def cov_p_w(self, hyp_dict, dom_size, D):
        """Sample w for maximum sampling"""
        raise NotImplementedError

    def base_multi(self, domain, hyp_vec, key_dict, hyp_dict):
        """Base covariance for multiple points.
        Used when calculating log probability (can't use
        stored values)
        Sum of parts"""
        temp_val = self.temp_kern.base_multi(
            np.array(domain[:,0], ndmin=2).T, 
            hyp_vec, key_dict[0], hyp_dict.values[0]
        )
        spat_val = self.spat_kern.base_multi(
            domain[:,1:], hyp_vec, key_dict[1], hyp_dict.values[1])
        
        if self.way == 'Summation':
            return temp_val + spat_val
        elif self.way == 'Multiplication':
            return temp_val * spat_val

    def k_cross_faster(self, domain, training, 
                       hyp_vec, key_dict, hyp_dict):
        temp_val = self.temp_kern.k_cross_faster(
            np.array(domain[:,0], ndmin=2).T,
            np.array(training[:,0], ndmin=2).T, 
            hyp_vec, key_dict[0], 
            hyp_dict.values[0])
        spat_val = self.spat_kern.k_cross_faster(
            domain[:,1:], training[:,1:], 
            hyp_vec, key_dict[1], 
            hyp_dict.values[1])
        if self.way == 'Summation':
            return temp_val + spat_val
        elif self.way == 'Multiplication':
            return temp_val * spat_val
    
    def covariance_partials(
        self, x1_pp, x2_pp, tau, hyp_dict, dd_dims=None
    ):
        """
        x1_pp: scalar, derivative-degree of point 1
        x2_pp: scalar, derivative-degree of point 2
        tau:   vector, distance between points
        hyp_dict: dict of hyperparameters
        dd_1:  scalar, first dimension for derivation
        dd_2:  scalar, second dimenion for derivation
        dd_3:  scalar, third dimension for derivation
        dd_4:  scalar, fourth dimenion for derivation
        
        Sum of kernel parts
        """
        if self.way == 'Multiplication':
            assert x1_pp == 0
            assert x2_pp == 0
            
        temp_val = self.temp_kern.covariance_partials(
            x1_pp, x2_pp, np.array([tau[0]]), 
            hyp_dict.values[0], dd_dims)
        spat_val = self.spat_kern.covariance_partials(
            x1_pp, x2_pp, tau[1:], 
            hyp_dict.values[1], dd_dims)
        if self.way == 'Summation':
            return temp_val + spat_val
        elif self.way == 'Multiplication':
            return temp_val * spat_val
        
    
def choose_kernel(kernel_name, parts=None, way='Summation'):
    if kernel_name == 'RBF':
        return RBF_kernel()
    elif kernel_name == 'Directed':
        return Directed_kernel()
    elif kernel_name == 'Sum':
        return Wind_sum_kernel()
    elif kernel_name == 'Product':
        return Wind_product_kernel()
    elif kernel_name == 'Combination':
        return Combination_kernel(parts=parts, way=way)
    elif kernel_name == 'Temporal':
        temp_kern = choose_kernel(parts[0])
        spat_kern = choose_kernel(parts[1])
        return Temporal_kernel(temp_kern=temp_kern,
                               spat_kern=spat_kern, way=way)
    else:
        raise ValueError
    
    
def product_normals(m1, m2, C1, C2):
    """Calculate product of two normal distributions"""
    try:
        C1_inv = np.linalg.inv(C1)
    except:
        C1_inv = np.linalg.inv(C1 + np.eye(len(C1)) * 0.0001)
    C2_inv = np.linalg.inv(C2)
    C_comb = np.linalg.inv(C1_inv + C2_inv)
    m_comb = np.matmul(
        C_comb, np.matmul(C1_inv, m1) + np.matmul(C2_inv, m2)
    )
    return (m_comb, C_comb)

