"""Module holding function to tune GP hyperparameters"""
import copy
import autograd.numpy as np
from autograd import grad
import scipy.optimize as scipy_opt
import autograd.scipy.stats as stat

def optimize_hyperparameters(
    kernel, xx, yy, hyp_dict, maxiter=1000, min_bb=0.001, 
    verbose=True, bound_dict=None, random_int=False,
    initialisations=1, prior_means=None, prior_stds=None
):
    """Function to optimize GP hyperparameters
    and hence tune the GP
    random_int: whether to randomly assign the first values to hyp_dict

    """

    print('Optimising with %s samples, %s initialisations' %(len(yy), initialisations))
    
    hyp_vec = copy.deepcopy(hyp_dict.array)

    log_prob_grad = grad(kernel.log_prob, 0)
    grad_f = lambda hh_f, xx_f, yy_f: log_prob_grad(
        hh_f, xx_f, yy_f, hyp_dict.key_dict, hyp_dict
    )


    fun_f = lambda hh_f, xx_f, yy_f: kernel.log_prob(
        hh_f, xx_f, yy_f, hyp_dict.key_dict, hyp_dict
    )
    
    if prior_stds is None:
        defaults = {
            'sigma_2': (min_bb, 10),
            'direc': (0.00000001, 2 * np.pi),
            'sigma_dir': (min_bb, 10),
            'll_dir': (0.05, 5),
            'sigma_rbf': (min_bb, 10),
            'll': (0.05, 5),
            'sigma_hat': (min_bb, 10)
        }

        
        if kernel.name not in ['Combination', 'Temporal']:
            if bound_dict is None:
                bound_dict = {}
            if set(bound_dict.keys()) != set(hyp_dict.key_dict.keys()):
                for key in set(hyp_dict.key_dict.keys()):
                    if key not in set(bound_dict.keys()):
                        bound_dict[key] = defaults[key]
                        if verbose:
                            print("Setting default %s: \t%s" 
                                  %(key, defaults[key]))   
        else:
            if bound_dict is None: 
                bound_dict = []
                for _ in range(len(hyp_dict.key_dict)):
                    bound_dict.append({})
            for kk in range(len(hyp_dict.key_dict)):
                if type(hyp_dict.key_dict[kk]) is dict:
                    if set(bound_dict[kk].keys()) != \
                            set(hyp_dict.key_dict[kk].keys()):
                        for key in set(hyp_dict.key_dict[kk].keys()):
                            if key not in set(bound_dict[kk].keys()):
                                bound_dict[kk][key] = defaults[key]
                                if verbose:
                                    print("%s: Setting default %s: \t%s" 
                                      %(kk, key, defaults[key]))  
                else:
                    for jj in range(len(hyp_dict.key_dict[kk])):
                        if set(bound_dict[kk][jj].keys()) != \
                                set(hyp_dict.key_dict[kk][jj]):
                            for key in set(hyp_dict.key_dict[kk][jj].keys()):
                                if key not in set(bound_dict[kk][jj].keys()):
                                    bound_dict[kk][jj][key] = defaults[key]
                                    if verbose:
                                        print("%s.%s: Setting default %s: \t%s" 
                                      %(kk, jj, key, defaults[key])) 

        if verbose:
            print(bound_dict)
        bounds_list = [[]]*hyp_dict.length
        if kernel.name not in ['Combination', 'Temporal']:
            for key in hyp_dict.key_dict.keys():
                bounds_list[hyp_dict.key_dict[key]] = bound_dict[key]
        else:
            for kk in range(len(hyp_dict.key_dict)):
                k_dict = hyp_dict.key_dict[kk]
                if type(k_dict) is dict:
                    for key in k_dict.keys():
                        bounds_list[k_dict[key]] \
                            = bound_dict[kk][key]
                else:
                    for jj in range(len(k_dict)):
                        k_dict_jj = k_dict[jj]
                        for key in k_dict_jj.keys():
                            bounds_list[k_dict_jj[key]] \
                                = bound_dict[kk][jj][key]
        
        bounds = tuple(bounds_list)
    elif prior_means is None:
        prior_means = np.zeros(len(hyp_dict.trunc_array(hyp_vec)))
    
    if prior_stds is not None:
        prior_f = lambda hh: stat.multivariate_normal.logpdf(
            np.log10(hyp_dict.trunc_array(hh)), 
            prior_means, 
            np.diag(prior_stds**2))
        grad_prior_f = grad(prior_f)
        opt_fun_f = lambda hh_f, xx_f, yy_f: fun_f(hh_f, xx_f, yy_f) - prior_f(hh_f)
        opt_grad_f = lambda hh_f, xx_f, yy_f: grad_f(hh_f, xx_f, yy_f) - grad_prior_f(hh_f)
        bounds = None
    else:
        opt_fun_f = fun_f
        opt_grad_f = grad_f
    
    if verbose:
        print('Gradient at start:')
        print(opt_grad_f(hyp_vec, xx, yy))
    
    log_probs = {}

    for ii in range(initialisations):
        if random_int or ii != 0:
            if kernel.name in ['Combination', 'Temporal']:
                key_dict_list = hyp_dict.key_dict
            else:
                key_dict_list = [hyp_dict.key_dict]
            for key_dict_sup in key_dict_list:
                if type(key_dict_sup) is dict:
                    key_dict_sup = [key_dict_sup]
                for key_dict in key_dict_sup:
                    for key in key_dict.keys():
                        idx = key_dict[key]
                        if prior_stds is None: 
                            if key == 'direc':
                                hyp_vec[idx] = np.random.uniform(np.pi, 2*np.pi)
                            else:
                                (bb_low, bb_high) = bounds[idx]

                                hyp_vec[idx] = np.exp(np.random.uniform(
                                    np.log(bb_low), np.log(bb_high)))

                        else:
                            if key == 'direc':
                                hyp_vec[idx] = np.random.uniform(0, np.pi)
                            else:
                                prior_idx = hyp_dict.prior_dict[key]
                                hyp_vec[idx] = 10**(np.random.normal(prior_means[prior_idx], 
                                                                 prior_stds[prior_idx]))


        print('Using initialisation %s' %hyp_vec)  
        print(bounds)
        result = scipy_opt.minimize(
            fun=opt_fun_f,
            x0=hyp_vec,
            args=(xx, yy),
            method="L-BFGS-B",
            options={"maxiter": maxiter},
            jac=opt_grad_f,
            bounds=bounds,
        )
        
        cur_log_prob = opt_fun_f(result.x, xx, yy)
        
        if verbose:
            print('-cur_log_prob:\t%s\t%s' %(-cur_log_prob, result.x))
        log_probs[cur_log_prob] = copy.deepcopy(result.x)
        
    
        
    min_log_prob = min(log_probs.keys())

    print('Log prob achieved: %s' %-min_log_prob)
    best_hyp_vec = log_probs[min_log_prob]

    hyp_dict.update_vals_from_array(best_hyp_vec)

    if prior_stds is None:
        if kernel.name not in ['Combination', 'Temporal']:
            for key in hyp_dict.key_dict.keys():
                if hyp_dict.get_val(key) in set(bound_dict[key]):
                    print('%s actively bounded' %key)
        else:
            for kk in range(len(hyp_dict.key_dict)):
                if type(hyp_dict.key_dict[kk]) is dict:
                    for key in hyp_dict.key_dict[kk].keys():
                        if hyp_dict.get_val(key, kk) in \
                                set(bound_dict[kk][key]):
                            print('%s-%s actively bounded' %(kk, key))
    
                else:
                    for jj in range(len(hyp_dict.key_dict[kk])):
                        for key in hyp_dict.key_dict[kk][jj].keys():
                            if hyp_dict.get_val(key, kk, jj) in \
                                    set(bound_dict[kk][jj][key]):
                                print('%s-%s-%s actively bounded' %(kk, jj, key))

    print(-opt_fun_f(hyp_dict.array, xx, yy))
    return hyp_dict


