"""Module for performing BO, related plotting and baseline
acquisition functions.
"""
import copy
import pickle
import time

import numpy as np
import scipy.stats as st

import matplotlib.pyplot as plt
from matplotlib.colors import to_rgba_array

try:
    from source import optimize_hyperparameters
except (ModuleNotFoundError, ImportError) as e:
    import optimize_hyperparameters


def acq_ucb(
    domain, xx_nn, yy_nn, kernel, hyp_dict, return_gp=False, log_folder=None,
    beta=1
):
    """Baseline upper confidence bound acquisition function
    log_folder is not used"""
    pred_mean, pred_var = kernel.pred_mean_var(
        domain, xx_nn, yy_nn, hyp_dict
    )
    pred_std = np.sqrt(pred_var)
    if return_gp:
        return (
            pred_mean + pred_std,
            pred_mean,
            pred_var,
            domain[np.argmax(pred_mean), :],
            domain[0, :],
        )
    else:
        return pred_mean + beta*pred_std


def acq_ucb_mean(pred_mean, pred_std, best_f=None, beta=1):
    """Baseline upper confidence bound acquisition function"""
    
    return pred_mean + beta*pred_std


def acq_PI(pred_mean, pred_std, best_f):
    """Baseline probability of improvement function.
    Needs signature updating."""
    return 1 - st.norm.cdf(best_f, loc=pred_mean, scale=pred_std)

def acq_weighted(pred_mean, pred_std, best_f, 
                 weights, base_acq):
    """Baseline probability of improvement function.
    Needs signature updating."""
    aa_vals = []
    for ii in range(len(weights)):
        aa_vals.append(base_acq(pred_mean[ii], 
                              pred_std[ii], 
                              best_f))
    aa_vals = np.array(aa_vals)
    W_norm = np.sum(weights)
    
    return (np.matmul(weights.T, aa_vals)/W_norm)[0,:]



def acq_EI(pred_mean, pred_std, best_f):
    """Baseline expected improvement function.
    beta parameter is not used.."""
    return (pred_mean - best_f) * acq_PI(
        pred_mean, pred_std, best_f
    ) + pred_std * st.norm.pdf(
        (best_f - pred_mean)/ pred_std
    )


def get_weighted_covariance(domain, xx_nn, yy_nn, kernel,
                           hyp_dict, hyp_samples):
    ND, _ = np.shape(domain)
    covariance = np.zeros((ND, ND))
    total_scores = 0
    
    for ii in range(hyp_samples['len']):
        arr_val = hyp_dict.array
        if hyp_dict.name not in ['Combination', 'Temporal']: 
            for key in hyp_dict.key_dict.keys():
                arr_val[hyp_dict.key_dict[key]] = hyp_samples[key][ii]
        else:
            for kk in range(len(hyp_dict.key_dict)):
                for key in hyp_dict.key_dict[kk].keys():
                    if key == 'sigma_2':
                        arr_val[hyp_dict.key_dict[kk][key]] = \
                            hyp_samples[key][ii]
                    else:
                        if key == 'sigma_rbf':
                            key_ts = 'sigma_r'
                        else:
                            key_ts = key
                        arr_val[hyp_dict.key_dict[kk][key]] = \
                            hyp_samples['%s_%s' %(kk, key_ts)][ii]
                    
        hyp_dict.update_vals_from_array(arr_val)
    
        cur_score = np.exp(-kernel.log_prob_cholesky(hyp_dict.array, 
                                       xx_nn, yy_nn, 
                                       hyp_dict.key_dict,
                                       hyp_dict))
        total_scores += cur_score
        cur_cov = kernel.kmm_extended_gp(domain, hyp_dict, m_deg=0) \
            + np.eye(ND)*hyp_dict.sigma_2
        covariance += cur_score * cur_cov
            
    covariance = covariance / total_scores
    return covariance

def get_weighted_post_sample(domain, xx_nn, yy_nn, kernel,
                            hyp_dict, hyp_samples):
    mean_vals = []
    var_vals = []
    scores = []
    
    for ii in range(hyp_samples['len']):
        arr_val = hyp_dict.array
        if hyp_dict.name not in ['Combination', 'Temporal']: 
            for key in hyp_dict.key_dict.keys():
                arr_val[hyp_dict.key_dict[key]] = hyp_samples[key][ii]
        else:
            for kk in range(len(hyp_dict.key_dict)):
                if type(hyp_dict.key_dict[kk]) is dict:
                    key_dict_sup = [hyp_dict.key_dict[kk]]
                else:
                    key_dict_sup = hyp_dict.key_dict[kk]
                for jj in range(len(key_dict_sup)):
                    for key in key_dict_sup[jj].keys():
                        if key == 'sigma_2':
                            arr_val[key_dict_sup[jj][key]] = \
                                hyp_samples[key][ii]
                        else:
                            if key in ['sigma_rbf', 'sigma_hat']:
                                key_ts = 'sigma_r'
                            elif key == 'direc':
                                key_ts = 'dir'
                            else:
                                key_ts = key
                            if len(key_dict_sup) > 1:
                                jj_str = '%s_' %jj
                            else:
                                jj_str = ''
                            arr_val[key_dict_sup[jj][key]] = \
                                hyp_samples['%s_%s%s' %(kk, jj_str, key_ts)][ii]
                    
        hyp_dict.update_vals_from_array(arr_val)
        pred_mean, pred_var = kernel.pred_internal_cholesky(
            domain, xx_nn, yy_nn, hyp_dict)

        mean_vals.append(pred_mean)
        var_vals.append(pred_var)
        scores.append(-kernel.log_prob_cholesky(hyp_dict.array, 
                                       xx_nn, yy_nn, 
                                       hyp_dict.key_dict,
                                       hyp_dict))
    cc = np.max(scores) - 10
    scores = np.array(np.exp(scores-cc), ndmin=2).T
    Z_scores = np.sum(scores)
    pred_mean = np.sum(mean_vals*scores,0)/Z_scores
    pred_var = np.sum([scores[ii]*var_vals[ii] 
                       for ii in range(len(scores))
                      ],0)/Z_scores

    return np.random.multivariate_normal(pred_mean, pred_var)


def get_weighted_pred(domain, xx_nn, yy_nn, kernel,
                           hyp_dict, hyp_samples,
                     return_parts=False):
    mean_vals = []
    std_vals = []
    scores = []
    
    for ii in range(hyp_samples['len']):
        arr_val = hyp_dict.array
        if hyp_dict.name not in ['Combination', 'Temporal']: 
            for key in hyp_dict.key_dict.keys():
                arr_val[hyp_dict.key_dict[key]] = hyp_samples[key][ii]
        else:
            for kk in range(len(hyp_dict.key_dict)):
                if type(hyp_dict.key_dict[kk]) is dict:
                    key_dict_sup = [hyp_dict.key_dict[kk]]
                else:
                    key_dict_sup = hyp_dict.key_dict[kk]
                for jj in range(len(key_dict_sup)):
                    for key in key_dict_sup[jj].keys():
                        if key == 'sigma_2':
                            arr_val[key_dict_sup[jj][key]] = \
                                hyp_samples[key][ii]
                        else:
                            if key in ['sigma_rbf', 'sigma_hat']:
                                key_ts = 'sigma_r'
                            elif key == 'direc':
                                key_ts = 'dir'
                            else:
                                key_ts = key
                            if len(key_dict_sup) > 1:
                                jj_str = '%s_' %jj
                            else:
                                jj_str = ''
                            arr_val[key_dict_sup[jj][key]] = \
                                hyp_samples['%s_%s%s' %(kk, jj_str, key_ts)][ii]
                    
        hyp_dict.update_vals_from_array(arr_val)
        pred_mean, pred_var = kernel.pred_mean_var_cholesky(
            domain, xx_nn, yy_nn, hyp_dict)
        if np.sum(np.isnan(pred_var)) > 0 or (np.sum(np.isnan(np.sqrt(pred_var))) > 0):
            print('Error: invalid value encountered')
        mean_vals.append(pred_mean)
        std_vals.append(np.sqrt(pred_var))
        scores.append(-kernel.log_prob_cholesky(hyp_dict.array, 
                                       xx_nn, yy_nn, 
                                       hyp_dict.key_dict,
                                       hyp_dict))
    cc = np.max(scores) - 10
    scores = np.array(np.exp(scores-cc), ndmin=2).T
    Z_scores = np.sum(scores)
    pred_mean = np.sum(mean_vals*scores,0)/Z_scores
    pred_std = np.sum(std_vals*scores,0)/Z_scores

    if return_parts:
        return mean_vals, std_vals, scores
    return pred_mean, pred_std

def optimize(
    Q,
    prob,
    acq_fnc,
    plot=True,
    plot_more=False,
    return_min=False,
    av_multiple=False,
    log_folder=None,
    bounds_dict=None,
    initialisations=1,
    prior_means=None,
    prior_stds=None,
    hyp_samples=None,
    verbose=False,
    use_weights=False
):
    """Function to perform BO, including tuning hyperparameters
    and sampling new points"""
    xx_nn = copy.deepcopy(prob.xx)
    yy_nn = copy.deepcopy(prob.yy)
    hyp_dict = copy.deepcopy(prob.hyp_dict)
    kernel = prob.kernel

    print('Sampling noise: %s' %prob.sampling_noise)
    if hyp_samples is not None:
        print(
            'Using acquisition function with %s hyperparameter samples.' %hyp_samples['len'])
        hyp_parameter_dev = None
    else:
        if prior_stds is not None:
            print("Using priors")
            print(prior_means)
            print(prior_stds)
            bounds_dict = None
        elif bounds_dict is not None:
            print("Using bounds:")
            print(bounds_dict)
        hyp_parameter_dev = np.zeros((len(hyp_dict.array), Q + 1))
        hyp_parameter_dev[:, 0] = hyp_dict.array
        
    if plot:
        _, axs = plt.subplots(Q, 2, figsize=(14, 5 * Q))
    maximums, minimums = [], []
    maximisers, minimisers = [], []
    

        
    if hyp_samples is not None:
        pred_mean, pred_std = get_weighted_pred(
                domain=prob.domain, xx_nn=xx_nn, yy_nn=yy_nn, 
                kernel=kernel, hyp_dict=hyp_dict, 
                hyp_samples=hyp_samples)
        if use_weights:
            pred_mean_list, pred_std_list, scores = get_weighted_pred(
                domain=prob.domain, xx_nn=xx_nn, yy_nn=yy_nn, 
                kernel=kernel, hyp_dict=hyp_dict, 
                hyp_samples=hyp_samples, 
                return_parts=True)
        
    for qq in range(Q):
        start_stage = time.time()
        if hyp_samples is None:
            hyp_dict = optimize_hyperparameters.optimize_hyperparameters(
                prob.kernel, xx_nn, yy_nn, hyp_dict, bound_dict=bounds_dict,
                initialisations=initialisations,
                prior_means=prior_means, prior_stds=prior_stds,
                verbose=verbose
            )
            if verbose: print('hyp_dict: \t%s' %hyp_dict.array)
            hyp_parameter_dev[:, qq + 1] = hyp_dict.array

            pred_mean, pred_var = kernel.pred_mean_var(
                        domain=prob.domain, X=xx_nn, y_nn=yy_nn, 
                        hyp_dict=hyp_dict
                        )
            pred_std = np.sqrt(pred_var)
            
        if plot_more or (log_folder is not None):

            target = np.max(yy_nn)
                
            if not use_weights:
                gg_acq = acq_fnc(
                    pred_mean=pred_mean, 
                    pred_std =pred_std, 
                    best_f=target)
            else:
                gg_acq = acq_fnc(
                    pred_mean=pred_mean_list, 
                    pred_std =pred_std_list, 
                    best_f=target,
                    weights=scores)
            x_opt, mu, var_mm, zz_mean = None, None, None, None

                
        else:
                
            target = np.max(yy_nn)
                
            if not use_weights:
                gg_acq = acq_fnc(
                    pred_mean=pred_mean, 
                    pred_std =pred_std, 
                    best_f=target)
            else:
                gg_acq = acq_fnc(
                    pred_mean=pred_mean_list, 
                    pred_std =pred_std_list, 
                    best_f=target,
                    weights=scores)
            x_opt, mu, var_mm, zz_mean = None, None, None, None

        if plot:
            plot_ext_model(
                    axs[qq, 1],
                    prob,
                    xx_nn,
                    yy_nn,
                    x_opt,
                    mu,
                    var_mm,
                    zz_mean,
                    av_multiple,
                    plot_more,
            )
            plot_acq_fun(gg_acq, prob, axs[qq, 0])
            
        if np.sum(np.isnan(gg_acq)) == len(prob.domain):
            print('Issue encountered.')
            print(xx_nn, yy_nn)
            print(gg_acq)
            print(pred_mean)
            print(pred_std)
            print(hyp_parameter_dev)
        sample = np.nanargmax(gg_acq)
        if np.sum(np.isnan(gg_acq)) != 0:
            print('Warning: %s nan values detected in iteration %s'
                 %(np.sum(np.isnan(gg_acq)), qq))
        xx_nn = np.vstack([xx_nn, prob.domain[sample]])
        yy_nn = np.hstack(
            [
                yy_nn,
                prob.true_func[sample]
                + np.random.normal(0, np.sqrt(prob.sampling_noise)),
            ]
        )
        
        if hyp_samples is None:
            pred_mean, _ = kernel.pred_mean_var(
                prob.domain, xx_nn, yy_nn, hyp_dict
            )
            hyp_array = hyp_dict.array
        else:
            pred_mean, pred_std = get_weighted_pred(
                prob.domain, xx_nn, yy_nn, kernel,
                               hyp_dict, hyp_samples)
            if use_weights:
                pred_mean_list, pred_std_list, scores = get_weighted_pred(
                prob.domain, xx_nn, yy_nn, kernel,
                               hyp_dict, hyp_samples,
                return_parts=True)
                
            hyp_array = None
        
        cur_guess = np.argmax(pred_mean)
        min_guess = np.argmin(pred_mean)
        maximums.append(pred_mean[cur_guess])
        maximisers.append(prob.domain[cur_guess])
        minimums.append(pred_mean[min_guess])
        minimisers.append(prob.domain[min_guess])
            
            
        time_stage = time.time() - start_stage

        if log_folder is not None:
            stage_dict = {
                "qq": qq,
                "gg_acq": gg_acq,
                "pred_mean": pred_mean,
                "x_opt": x_opt,
                "mu": mu,
                "var_mm": var_mm,
                "zz_mean": zz_mean,
                "time_stage": time_stage,
                "hyp_array": hyp_array,
                "xx_nn": xx_nn,
                "yy_nn": yy_nn,
                "stage_started": start_stage,
            }
            pickle.dump(
                stage_dict,
                open(log_folder + ("stage_%s.p" % qq), "wb"),
            )
        

        if plot:
            plot_new_sample(axs[qq, 1], prob, xx_nn, yy_nn)
            plot_add_guesses(
                axs[qq, 1], prob, cur_guess, min_guess, 
                pred_mean, plot_min=False
            )


    if return_min:
        return (
            maximums,
            maximisers,
            minimums,
            minimisers,
            hyp_parameter_dev,
            xx_nn,
            yy_nn,
        )
    else:
        return (
            maximums,
            maximisers,
            hyp_parameter_dev,
            xx_nn,
            yy_nn,
        )


def plot_hyperparameters(key_dict, hyp_parameter_dev, ax):
    """Plot hyperparameter development"""
    for key in key_dict.keys():
        ax.plot(
            hyp_parameter_dev[key_dict[key], :], "-x", label=key
        )
    ax.grid()
    ax.legend()


def plot_progress_optimum(optimisers, ax, optimum, prob):
    """Plot the progress of minimisers/maximisers towards
    the optimum."""
    Q = len(optimisers)
    opt_array = np.array(optimisers)
    D = len(optimum)

    if D == 1:
        ax.scatter(list(range(Q)), optimisers)
        ax.plot([0, Q], [optimum] * 2)
    elif D == 2:
        ax.scatter(
            opt_array[:, 0] + 0.005 * np.random.randn(Q),
            opt_array[:, 1] + 0.005 * np.random.randn(Q),
            marker="x",
            c=list(range(Q)),
            vmin=0,
            vmax=Q,
            s=100,
            alpha=1,
            cmap="Greens",
        )
        ax.scatter(
            opt_array[-1, 0],
            opt_array[-1, 1],
            s=150,
            c="green",
            marker="x",
        )
        ax.plot(
            opt_array[:, 0],
            opt_array[:, 1],
            "green",
            ":",
            alpha=0.5,
        )
        ax.scatter(optimum[0], optimum[1], s=100, edgecolors="k")
        x_min, x_max = prob.dom_1D_vecs[0][0], prob.dom_1D_vecs[0][-1]
        y_min, y_max = prob.dom_1D_vecs[1][0], prob.dom_1D_vecs[1][-1]
        x_epsilon = 0.05*(x_max-x_min)
        y_epsilon = 0.05*(y_max-y_min)
        
        ax.axis([x_min-x_epsilon, x_max+x_epsilon, y_min-y_epsilon, y_max+y_epsilon])


def plot_optimisation(
    maximums, maximisers, minimums, minimisers, prob
):
    """Plot development of maximisers, minimisers,
    minimums and maximums"""
    Q = len(maximisers)
    _, axs = plt.subplots(1, 4, figsize=(16, 4))
    axs[0].plot(maximums)
    axs[0].plot([0, Q - 1], [prob.maximum] * 2)
    plot_progress_optimum(maximisers, axs[1], prob.maximiser, prob)
    axs[2].plot(minimums)
    axs[2].plot([0, Q - 1], [prob.minimum] * 2)
    plot_progress_optimum(minimisers, axs[3], prob.minimiser, prob)
    axs[0].set_title("Maximum")
    axs[1].set_title("Maximiser")
    axs[2].set_title("Minimum")
    axs[3].set_title("Minimiser")
    axs[0].set_xlabel("Iteration")
    axs[1].set_xlabel("Iteration")
    axs[2].set_xlabel("Iteration")
    axs[3].set_xlabel("Iteration")


def plot_acq_fun(gg_acq, prob, ax):
    """Plot acquisition at one stage"""
    if prob.D == 1:
        ax.plot(prob.domain, gg_acq)
        ax.set_xlim([-0.02, 1.02])
        ax.grid()
    elif prob.D == 2:
        x_axes = prob.shifted_axes()
        ax.pcolormesh(
            x_axes[0],
            x_axes[1],
            np.reshape(
                gg_acq,
                (prob.dom_lengths[0], prob.dom_lengths[1]),
            ).T,
            cmap="cividis",
            alpha=0.8,
        )


def plot_new_sample(ax, prob, xx, yy):
    """Add the new sample to domain plot"""
    if prob.D == 1:
        ax.scatter(xx[-1], yy[-1], label="next sample")
    elif prob.D == 2:
        cc = get_hex_col(
            yy[-1], 0, prob.minimum, prob.maximum, 1
        )
        ax.scatter(
            [xx[-1, 0]],
            [xx[-1, 1]],
            c=cc,
            vmin=prob.minimum,
            vmax=prob.maximum,
            edgecolors="k",
            s=200,
        )


def plot_add_guesses(ax, prob, cur_guess, min_guess, pred_mean,
                    plot_min=True):
    """Add guesses for minimiser and maximiser to domain plot"""
    if prob.D == 1:
        if plot_min:
            locs = [prob.domain[cur_guess], prob.domain[min_guess]]
            vals = [pred_mean[cur_guess], pred_mean[min_guess]]
        else:
            locs = [prob.domain[cur_guess]]
            vals = [pred_mean[cur_guess]]
        ax.scatter(
            locs,
            vals,
            label="cur guesses",
        )
        ax.legend()
    elif prob.D == 2:
        if plot_min:
            locs_0 = [
                prob.domain[cur_guess, 0],
                prob.domain[min_guess, 0],
            ]
            locs_1 = [
                prob.domain[cur_guess, 1],
                prob.domain[min_guess, 1],
            ]
            vals = [
                pred_mean[cur_guess],
                pred_mean[min_guess],
            ]
        else:
            locs_0 = [prob.domain[cur_guess, 0]]
            locs_1 = [prob.domain[cur_guess, 1]]
            vals = [pred_mean[cur_guess]]
        cc = [
            get_hex_col(ii, 0, prob.minimum, prob.maximum, 1)
            for ii in vals
        ]
        ax.scatter(
            locs_0,
            locs_1,
            c=cc,
            edgecolors="k",
            linewidths=2,
            s=100,
        )
        ax.scatter(prob.maximiser[0],
                      prob.maximiser[1],
                      marker='x',
                      c='k')


def plot_ext_model(
    ax,
    prob,
    xx,
    yy,
    x_opt,
    mu,
    var_mm,
    zz_mean,
    av_multiple,
    plot_more,
    plot_samples=False
):
    """Plot current state of model."""
    if prob.D == 1:
        ax.scatter(xx, yy, label="samples")
        ax.set_xlim([-0.02, 1.02])
        ax.plot(prob.domain, prob.true_func, alpha=0.5)
        ax.grid()
        if plot_more:
            if av_multiple:
                for aa in range(len(x_opt)):
                    ax.plot(prob.domain, mu[aa], color="C1")
                    ax.fill_between(
                        prob.domain[:, 0],
                        (mu[aa] - np.sqrt(var_mm[aa])),
                        (mu[aa] + np.sqrt(var_mm[aa])),
                        color="C1",
                        alpha=0.1,
                    )
                ax.scatter(
                    x_opt[0],
                    zz_mean[0],
                    color="C1",
                    label="x_opt",
                )
                for aa in range(1, len(x_opt)):
                    ax.scatter(
                        x_opt[aa], zz_mean[aa], color="C1"
                    )
            else:
                ax.plot(prob.domain, mu)
                ax.fill_between(
                    prob.domain[:, 0],
                    (mu - np.sqrt(var_mm)),
                    (mu + np.sqrt(var_mm)),
                    color="C1",
                    alpha=0.4,
                )
                ax.scatter(x_opt, zz_mean, label="x_opt")
    elif prob.D == 2:

        if plot_more and not av_multiple:
            std_mm = np.sqrt(var_mm)
            M = max(std_mm)
            colours = []
            for ii in range(prob.dom_size):
                colours.append(
                    get_hex_col(
                        mu[ii],
                        std_mm[ii],
                        prob.minimum,
                        prob.maximum,
                        M,
                    )
                )
            ax.imshow(
                prob.reshape_array_pixels(
                    to_rgba_array(colours)
                ),
                origin="lower",
                extent=[
                    -prob.dom_steps[0] / 2,
                    1 + prob.dom_steps[0] / 2,
                    -prob.dom_steps[1] / 2,
                    1 + prob.dom_steps[1] / 2,
                ],

            )
            if plot_samples:
                samp_cols = [] 
                for ii in range(len(yy)):
                    samp_cols.append(
                        get_hex_col(
                            yy[ii],
                            0,
                            prob.minimum,
                            prob.maximum,
                            M,
                        )
                    )
                
                ax.scatter(
                    xx[:,0],
                    xx[:,1],
                    c=samp_cols,
                    edgecolors="k",
                )
        else:
            prob.plot_2D(
                ax,
                prob.true_func,
                prob.minimum,
                prob.maximum,
                xx,
                yy,
            )
            



def conv_hex(flo):
    """Convert colout to hex"""
    return format(int(255 * flo), "02x")


def get_hex_col(val, conf, vmin, vmax, M):
    """Convert values to colour codes"""
    if M == 0:
        M = 1
    mean = (vmax - vmin) / 2 + vmin
    if val < vmin:
        val = vmin
    if val > vmax:
        val = vmax
    if val < mean:
        r_val = (val - vmin) / (
            mean - vmin
        )  # Reduce if value is below mean
        b_val = 1
    else:
        r_val = 1
        b_val = (vmax - val) / (
            vmax - mean
        )  # Reduce if value is above mean
    alph_g = conf / M
    alph_r = (1 - r_val) * conf / M
    alph_b = (1 - b_val) * conf / M
    return (
        "#"
        + conv_hex(r_val + alph_r)
        + conv_hex(alph_g)
        + conv_hex(b_val + alph_b)
    )
