"""Module to setup problems, and problem class"""
import numpy as np
from matplotlib.colors import to_rgba_array, ListedColormap
import matplotlib.pyplot as plt


try:
    from source import kernel_functions_mul as kernel_functions
    from source import outer_loop as ol
    from source import hyp_dict as hyp_dict_module
except:
    import kernel_functions_mul as kernel_functions
    import outer_loop as ol
    import hyp_dict as hyp_dict_module

class Problem_obj:
    """Class to hold BO problems"""

    def __init__(
        self,
        xx_nn,
        yy_nn,
        domain,
        true_func,
        kernel,
        hyp_dict,
        C,
        dom_1D_vecs,
        seed,
        sampling_noise=0.001
    ):
        self.xx = xx_nn
        self.yy = yy_nn
        self.domain = domain
        self.dom_size, self.D = np.shape(domain)
        self.dom_1D_vecs = dom_1D_vecs
        if dom_1D_vecs is not None:
            self.dom_lengths = [len(dd) for dd in dom_1D_vecs]
            self.dom_steps = [
                dd[1] - dd[0] for dd in dom_1D_vecs
            ]
            self.dom_bounds = np.array([
                [dom_1D_vecs[0,0], dom_1D_vecs[0,-1]],
                [dom_1D_vecs[1,0], dom_1D_vecs[1,-1]]
            ])
        self.true_func = true_func
        self.maximum = np.max(true_func)
        self.maximiser = domain[np.argmax(true_func)]
        self.minimum = np.min(true_func)
        self.minimiser = domain[np.argmin(true_func)]
        self.hyp_dict = hyp_dict
        self.kernel = kernel
        self.C = C
        self.seed = seed
        self.sampling_noise = sampling_noise

    def shifted_axes(self):
        """Function to help handle shifted axes"""
        x_axes = []
        assert self.D == 2
        for ii in range(2):
            xii_axs = (
                self.dom_1D_vecs[ii] - self.dom_steps[ii] * 0.5
            )
            xii_axs = np.hstack(
                [xii_axs, [xii_axs[-1] + self.dom_steps[ii]]]
            )
            x_axes.append(xii_axs)
        return x_axes

    def reshape_array_2D(self, vec):
        """Function to reshape the array into 2D"""
        assert self.D == 2
        return np.reshape(
            vec, (self.dom_lengths[0], self.dom_lengths[1])
        ).T

    def reshape_array_pixels(self, vec):
        """Function to reshape pixel array"""
        return np.transpose(
            np.reshape(
                np.array(vec),
                (self.dom_lengths[0], self.dom_lengths[1], 4),
            ),
            (1, 0, 2),
        )

    def plot_2D(self, ax, values, vmin, vmax, xx, yy, show_colorbar=False):
        """Plot problem specification for 2D problems"""
        c_scat = [
            ol.get_hex_col(cc, 0, vmin, vmax, 1) for cc in yy
        ]
        c_true = [
            ol.get_hex_col(cc, 0, vmin, vmax, 1) for cc in values
        ]
        
        if show_colorbar:
            N = 100
            range_array = np.linspace(vmin, vmax, N)
            hex_array = [ol.get_hex_col(cc, 0, vmin, vmax, 1) for cc in range_array]
            rgba_array = to_rgba_array(hex_array)
            rgba_array = np.reshape(rgba_array, (N, 1, 4))

            raw_cmap = np.c_[rgba_array[:,:,0],rgba_array[:,:,1],rgba_array[:,:,2],rgba_array[:,:,3]]
            cmap = ListedColormap(raw_cmap)
        
            im = ax.imshow(
                self.reshape_array_pixels(to_rgba_array(c_true)),
                origin="lower",
                extent=[
                    self.dom_bounds[0,0] - self.dom_steps[0] / 2,
                    self.dom_bounds[0,1] + self.dom_steps[0] / 2,
                    self.dom_bounds[1,0] - self.dom_steps[1] / 2,
                    self.dom_bounds[1,1] + self.dom_steps[1] / 2,
                ],
                aspect="auto",
                cmap=cmap
            )
        
            cbar = plt.colorbar(im, ax=ax)
            num_tick_labels = len(list(cbar.ax.get_yticklabels()))
            ylabels = np.round(np.linspace(vmin, vmax, num_tick_labels), 2)
            cbar.ax.set_yticklabels(ylabels);
        else:
            im = ax.imshow(
                self.reshape_array_pixels(to_rgba_array(c_true)),
                origin="lower",
                extent=[
                    self.dom_bounds[0,0] - self.dom_steps[0] / 2,
                    self.dom_bounds[0,1] + self.dom_steps[0] / 2,
                    self.dom_bounds[1,0] - self.dom_steps[1] / 2,
                    self.dom_bounds[1,1] + self.dom_steps[1] / 2,
                ],
                aspect="auto"
            )
        
        ax.scatter(
            xx[:, 0], xx[:, 1], c=c_scat, edgecolors="k", s=100
        )

    def plot(self, ax, alpha=1, xx=None, yy=None, values=None, show_colorbar=False):
        """Plot problem specification"""
        if xx is None or yy is None:
            xx, yy = self.xx, self.yy
        if values is None:
            values = self.true_func
        if self.D == 1:
            ax.plot(self.domain, values, alpha=alpha)
            ax.scatter(xx, yy, alpha=alpha)
            ax.scatter(
                [self.maximiser, self.minimiser],
                [self.maximum, self.minimum],
                alpha=alpha,
            )
        elif self.D == 2:
            mappable = self.plot_2D(
                ax,
                values,
                self.minimum,
                self.maximum,
                xx,
                yy,
                show_colorbar=show_colorbar
            )
        else:
            raise NotImplementedError

    def print_properties(self):
        """Print some problem properties"""
        print("llscale:\t%s" % self.hyp_dict.ll)
        print("sigma_2:\t%s" % self.hyp_dict.sigma_2)


def problem_a(num_samples, prob_seed=13, llscale=1):
    """Build a simple 1D example problem"""
    dom_size = 101
    domain = np.reshape(
        np.linspace(0, 1, dom_size), (dom_size, 1)
    )

    sigma_2 = 0.001

    kernel = kernel_functions.RBF_kernel()
    llscale_array = np.array([llscale])
    hyp_dict = hyp_dict_module.Hyp_dict(
        {
            "ll": llscale_array,
            "sigma_rbf": 1,
            "sigma_2": sigma_2,
            "sigma_n": np.sqrt(sigma_2),
            "kernel_name": 'RBF'
        }
    )

    C = np.zeros((dom_size, dom_size))
    for ii in range(dom_size):
        for jj in range(dom_size):
            C[ii, jj] = kernel.covariance_partials(
                0, 0, domain[ii, :] - domain[jj, :], hyp_dict
            )

    np.random.seed(prob_seed)
    true_func = np.random.multivariate_normal(
        np.zeros(dom_size), C
    )

    np.random.seed(prob_seed)
    sampled_indices = np.random.randint(0, dom_size, num_samples)

    xx_nn = domain[sampled_indices, :]
    yy_nn = true_func[sampled_indices] + np.random.normal(
        0, np.sqrt(sigma_2), num_samples
    )

    prob = Problem_obj(
        xx_nn,
        yy_nn,
        domain,
        true_func,
        kernel,
        hyp_dict,
        C,
        np.array([domain], ndmin=2).T,
        prob_seed,
    )
    return prob


def problem_multi(
    num_samples,
    prob_seed=13,
    llscale=[1, 1],
    num_dim=2,
    dom_size_1D=11,
    kernel_type="RBF",
    sigma_rbf=1,
    sigma_dir=1,
    sigma_hat=1,
    sigma_2=0.001,
    gamma=None,
    dom_start=0,
    dom_end=1
):
    """Build a simple multi-dimensional example"""

    dom_1D = np.linspace(dom_start, dom_end, dom_size_1D)

    all_p = list([aa] for aa in dom_1D)
    for _ in range(num_dim - 1):
        new_all_p = []
        for pp in all_p:
            for vv in list(dom_1D):
                new_all_p.append(pp + [vv])
        all_p = new_all_p
    domain = np.array(all_p)
    dom_size, _ = np.shape(domain)

    if gamma is None:
        direc = np.array([1, 0.5])
    else:
        direc = np.array([np.cos(gamma), np.sin(gamma)])
    
    if kernel_type == "RBF":
        kernel = kernel_functions.RBF_kernel()
        llscale_array = np.array(llscale)
        hyp_dict = hyp_dict_module.Hyp_dict(
            {
                "ll": np.array([llscale[0]] * 2),
                "sigma_rbf": sigma_rbf,
                "sigma_2": sigma_2,
                "sigma_n": np.sqrt(sigma_2),
                "kernel_name": "RBF"
            }
        )
    elif kernel_type == "Sum":
        assert num_dim == 2
        kernel = kernel_functions.Wind_sum_kernel()

        hyp_dict = hyp_dict_module.Hyp_dict(
            {
                "ll": np.array([llscale[0]] * 2),
                "sigma_rbf": sigma_rbf,
                "sigma_dir": sigma_dir,
                "ll_dir": llscale[1],
                "direc": direc,
                "sigma_2": sigma_2,
                "sigma_n": np.sqrt(sigma_2),
                "kernel_name": 'Sum'
            }
        )
    elif kernel_type == "Product":
        assert num_dim == 2
        kernel = kernel_functions.Wind_product_kernel()
        hyp_dict = hyp_dict_module.Hyp_dict(
            {
                "ll": np.array([llscale[0]] * 2),
                "sigma_rbf": sigma_hat,
                "ll_dir": llscale[1],
                "direc": direc,
                "sigma_2": sigma_2,
                "sigma_n": np.sqrt(sigma_2),
                "kernel_name": 'Product'
            }
        )
    elif kernel_type == "Directed":
        assert num_dim == 2
        kernel = kernel_functions.Directed_kernel()
        hyp_dict = hyp_dict_module.Hyp_dict(
            {
                "sigma_dir": sigma_dir,
                "ll_dir": llscale[1],
                "direc": direc,
                "sigma_2": sigma_2,
                "sigma_n": np.sqrt(sigma_2),
                "kernel_name": 'Directed'
            }
        )
    else:
        raise ValueError

    C = np.zeros((dom_size, dom_size))
    for ii in range(dom_size):
        for jj in range(dom_size):
            C[ii, jj] = kernel.covariance_partials(
                0, 0, domain[ii, :] - domain[jj, :], hyp_dict
            )

    np.random.seed(prob_seed)
    true_func = np.random.multivariate_normal(
        np.zeros(dom_size), C
    )

    np.random.seed(prob_seed)
    sampled_indices = np.random.randint(0, dom_size, num_samples)

    xx_nn = domain[sampled_indices, :]
    yy_nn = true_func[sampled_indices] + np.random.normal(
        0, np.sqrt(sigma_2), num_samples
    )

    dom_1D_vecs = np.array([dom_1D] * 2)

    prob = Problem_obj(
        xx_nn,
        yy_nn,
        domain,
        true_func,
        kernel,
        hyp_dict,
        C,
        dom_1D_vecs,
        prob_seed,
    )
    return prob
