
import numpy as np
import copy

def resample_theta_multiple(
    tt, likelihood_tt, eta, xx_grid, sample, sample_theta_func, 
    calc_log_likelihood):
    # Sample {theta_i} | eta
    # Proposal is p(theta_i|eta)
    # Pi is p(D_i|theta_i)p(theta_i|eta)
    # p(theta_i|eta) terms cancel out
    
    new_tts = []
    new_tt_likelihoods = []
    for ii in range(len(tt)):
        tt_ii, likelihood_tt_ii = resample_theta(
            tt[ii], likelihood_tt[ii], eta, [xx_grid[ii]], 
            [sample[ii]], 
            sample_theta_func, calc_log_likelihood)
    
        new_tts.append(tt_ii)
        new_tt_likelihoods.append(likelihood_tt_ii)
    return np.array(new_tts), np.array(new_tt_likelihoods)

def resample_theta(
    tt, likelihood_tt, eta, xx_grid, sample, sample_theta_func, 
    calc_log_likelihood):
    # Sample theta | eta
    # Proposal is p(theta|eta)
    # Pi is p(D|theta)p(theta|eta)
    # p(theta|eta) terms cancel out
    
    proposal = sample_theta_func(eta)
    likelihood_proposal = calc_log_likelihood(proposal, xx_grid, sample)
    
    aa = min(0, likelihood_proposal - likelihood_tt)
    if np.log(np.random.uniform(0, 1)) <= aa:
        return proposal, likelihood_proposal
    else:
        return tt, likelihood_tt
    
def resample_eta_by_parts(theta, ee, log_pdf_theta_func):
    cur_ee = copy.deepcopy(ee)
    for ii in range(len(ee.array())):
        cur_ee, cur_lik = resample_eta(theta, cur_ee, 
                              log_pdf_theta_func,
                             prop_part=ii)
    return cur_ee, cur_lik

def resample_eta(theta, ee, log_pdf_theta_func, 
                 prop_part=None):
    # Sample eta | theta
    # Proposal is a Normal around previous eta
    # Pi is p(eta|theta) which is proportional to p(theta|eta)
    if prop_part is None:
        proposal = ee.propose()
    else:
        proposal = ee.propose_part(ii=prop_part)
    
    likelihood_ee = log_pdf_theta_func(theta, ee) \
                    + np.log(ee.prior())
    prior_prop = np.log(proposal.prior())
    likelihood_proposal = log_pdf_theta_func(theta, proposal) \
                    + prior_prop
    print(np.log(ee.prior()), log_pdf_theta_func(theta, ee),
        np.log(proposal.prior()), log_pdf_theta_func(theta, proposal))
    
    aa = min(0, likelihood_proposal - likelihood_ee)
    
    if (prior_prop > -np.inf) and \
            (np.log(np.random.uniform(0, 1)) <= aa):
        return proposal, likelihood_proposal
    else:
        return ee, likelihood_ee
    
def eta_part(prob, params, part, SCALE_GLOBAL):
    if part == 'lscale':
        return prob.eta_vars(SCALE_GLOBAL,
                            params[0], params[1], SCALE_GLOBAL)
    if part == 'sigma_n':
        return prob.eta_vars(SCALE_GLOBAL,
                1, 1, SCALE_GLOBAL,
                sig_n_aa=params[0], sig_n_loc=params[1], sig_n_scale=SCALE_GLOBAL)
    if part == 'sigma_r':
        return prob.eta_vars(SCALE_GLOBAL,
                1, 1, SCALE_GLOBAL,
                sig_r_aa=params[0], sig_r_loc=params[1], sig_r_scale=SCALE_GLOBAL)

def get_prob_overlap(prob, eta, eta_samples, burn_in, N_samp, 
                      SCALE_GLOBAL, part, B=1000, return_samples=False, 
                     start_grid=0, end_grid=6):

    x_grid = np.linspace(start_grid, end_grid, B)
    est_prob = np.zeros(B)
    for theta in eta_samples[burn_in:,:]:
        eta_tmp = eta_part(prob, theta, part, SCALE_GLOBAL)
        est_prob += prob.theta_part_pdf(x_grid, eta_tmp, part=part)
    est_prob = est_prob/(N_samp-burn_in) # Normalise
    true_prob = prob.theta_part_pdf(x_grid, eta, part=part)
    prob_overlap = np.sum(np.minimum(est_prob, true_prob))*(end_grid-start_grid)/B
    if return_samples:
        return prob_overlap, x_grid, est_prob, true_prob
    return prob_overlap

def get_samples(prob, part, B, sample_P, eta, 
                N_samp=None, burn_in=None, 
                eta_samples=None):
    if part == 'lscale':
        samp_theta = lambda ee: prob.sample_theta(ee).lscale
    elif part == 'sigma_n':
        samp_theta = lambda ee: prob.sample_theta(ee).sigma_n
    elif part == 'sigma_r':
        samp_theta = lambda ee: prob.sample_theta(ee).sigma_r
        
    if sample_P: 
        KL_samples = np.array(
            [samp_theta(eta) for _ in range(B)])
    else:
        KL_samples = []
        for _ in range(B):
            rand_idx = np.random.randint(0, N_samp-burn_in)
            eta_params = eta_samples[burn_in+rand_idx,:]
            eta_tmp = eta.eta_vars_array(eta_params)
            KL_samples.append(samp_theta(eta_tmp)               )
        KL_samples = np.array(KL_samples)
    return KL_samples

def get_KL_divergence(prob, eta, eta_samples, burn_in, N_samp, 
                      SCALE_GLOBAL, part, B=1000, sample_P = True,
                      return_samples=False, jensen_half=False):

    KL_samples = get_samples(prob, part, B, sample_P, eta, N_samp,
                            burn_in, eta_samples)

    est_prob = np.zeros(B)
    for theta in eta_samples[burn_in:,:]:
        eta_tmp = eta_part(prob, theta, part, SCALE_GLOBAL)
        est_prob += prob.theta_part_pdf(
            KL_samples, eta_tmp, part=part)
    est_prob = est_prob/(N_samp-burn_in) # Normalise
    true_prob = prob.theta_part_pdf(KL_samples, eta, part=part)
    
    if sample_P:
        if jensen_half:
            KL_div = np.mean(np.log(true_prob/
                    (0.5*est_prob+0.5*true_prob)))
        else:
            KL_div = np.mean(np.log(true_prob/est_prob))
    else:
        if jensen_half:
            KL_div = np.mean(np.log(est_prob/
                     (0.5*est_prob+0.5*true_prob)))

        else:
            KL_div = np.mean(np.log(est_prob/true_prob))

    if return_samples:
        return KL_div, KL_samples, est_prob, true_prob
    return KL_div

def get_JS_divergence(prob, eta, eta_samples, burn_in, N_samp, 
                      SCALE_GLOBAL, part, B=1000, return_samples=False):
    
    partP = get_KL_divergence(
            prob, eta, eta_samples, burn_in, N_samp,
            SCALE_GLOBAL, part, B, sample_P=True, 
            return_samples=return_samples, jensen_half=True)
    partQ = get_KL_divergence(
            prob, eta, eta_samples, burn_in, N_samp,
            SCALE_GLOBAL, part, B, sample_P=False, 
            return_samples=return_samples, jensen_half=True)
    
    if not return_samples:
        return 0.5*(partP+partQ)
    else:
        JS_div = 0.5*(partP[0]+partQ[0])
        JS_samples = np.vstack([partP[1], partQ[1]])
        JS_est_prob = np.vstack([partP[2], partQ[2]])
        JS_true_prob = np.vstack([partP[3], partQ[3]])
        return JS_div, JS_samples, JS_est_prob, JS_true_prob
