import MCMC
import problems

import time
import datetime
import pickle
import numpy as np

import argparse
from pathlib import Path


import sys
try:
    sys.path.append('../../BO/')
    sys.path.append('../../BO/source/')
    import source
except:
    sys.path.append('../../BO/')
    sys.path.append('../../BO/source/')
    import source
import kernel_functions_mul
import hyp_dict as hd
    
def parse_arguments():
    parser = argparse.ArgumentParser()
    
    parser.add_argument('-R', help='number of training images', type=int)
    parser.add_argument('-N', help='number of points per training image', type=int)
    parser.add_argument('-N_samp', help='number of MCMC samples', type=int)
    parser.add_argument('-M', help='number of times to sample eta per theta resampling', type=int)
    parser.add_argument('-burn_in', help='# MCMC burn in samples', type=int)
    parser.add_argument('-ll_init', help='initial values eta lscale', type=float,
                        nargs=2)
    parser.add_argument('-ll_dir_init', help='initial values eta directional lscale', 
                        type=float, nargs=2, default=None)
    parser.add_argument('-sigma_n_init', help='initial values eta sigma_n', 
                        type=float, nargs=2)
    parser.add_argument('-sigma_r_init', help='initial values eta sigma_r', 
                        type=float, nargs=2)
    parser.add_argument('-sigma_dir_init', help='initial values eta sigma_d', 
                        type=float, nargs=2, default=None)
    parser.add_argument('-ll_true', help='true values eta lscale', type=float,
                        nargs=2)
    parser.add_argument('-ll_dir_true', help='true values eta directional lscale', 
                        type=float, nargs=2, default=None)
    parser.add_argument('-sigma_n_true', help='true values eta sigma_n', type=float,
                        nargs=2)
    parser.add_argument('-sigma_dir_true', help='true values eta sigma_dir', type=float,
                        nargs=2, default=None)
    parser.add_argument('-sigma_r_true', help='true values eta sigma_r', type=float,
                        nargs=2)
    parser.add_argument('-folder', help='folder to put results in', type=str, default=None)
    parser.add_argument('-local', help='whether the script is being run locally', type=int, default=1)
    parser.add_argument('-kernel', help='covariance kernel to use', type=str)
    parser.add_argument('-ll_prop', help='std when proposing ll', type=float,
                       default=(1.5, 0.5), nargs=2)
    parser.add_argument('-ll_dir_prop', help='std when proposing ll_dir', 
                        type=float,
                       default=(1.5, 0.5), nargs=2)
    parser.add_argument('-sigma_n_prop', help='std when proposing sigma_n', 
                        type=float,
                       default=(0.3, 0.1), nargs=2)
    parser.add_argument('-sigma_dir_prop', help='std when proposing sigma_d', 
                        type=float,
                       default=(0.3, 0.1), nargs=2)
    parser.add_argument('-sigma_r_prop', help='std when proposing sigma_r', 
                        type=float,
                       default=(0.3, 0.1), nargs=2)
    parser.add_argument('-data_locs', help='first part of files to load data from, followed by first and last id to use',
                        type=str, default=None, nargs=3)
    parser.add_argument('--parts', help='when combining multiple kernels',
                        type=str, default=None, nargs='+')
    parser.add_argument('-constant', help='hyp distributions to be kept constant.', type=str, default=['dir'], nargs='+')
    parser.add_argument('-kernel_hyp_file', help='location of pickled tuple with kernel and dict with correct structure.', type=str, default=None)

    args = parser.parse_args()
    
    return args


def pickle_state(ee_samples, tt_samples, args, samples, locs, proposals,
                 timestamp, log_folder):
       
    ee_arrays = np.array([ee_samp.array() for ee_samp in ee_samples])
    to_pickle = {'ee_arrays': ee_arrays, 'args': args,
                'samples_data': samples,
                'locs': locs, 'order': ee.order,
                'constants': ee.constant}
    
    # Build arrays of samples
    key_translator = {'ll': 'lscales',
                  'sigma_n': 'noise_sigmas',
                  'sigma_r': 'kern_sigmas',
                  'll_dir': 'll_dirs',
                  'sigma_dir': 'sigma_dirs',
                  'dir': 'dirs'}
    for key in ee_true.order:
        if args.kernel in ['Combination', 'Temporal'] and key != 'sigma_n':
            splt = key.find('_')
            splt_two = key[splt+1:].find('_')
            if key[splt+1:splt_two+splt+1].isnumeric():
                kk_str = '%s_' %key[:splt_two+splt+1]
                key_part = key[splt+splt_two+2:]

            else:
                kk_str = '%s_' %key[:splt]
                key_part = key[splt+1:]
        else:
            kk_str = ''
            key_part = key
            
        key_list = [[tt.vars[key] for tt in tt_samp]
                    for tt_samp in tt_samples]
        to_pickle['%s%s'%(kk_str, key_translator[key_part])] = key_list
    
            
    proposals = np.array(proposals)

    for ii in range(len(tt_samples[0][0].order)):
        key = tt_samples[0][0].order[ii]
        if key == 'll':
            string = 'lambdas_true'
        elif key == 'sigma_rbf':
            string = 'sigma_rs_true'
        else:
            string = key+'s_true'
        if args.data_locs is None:
            to_pickle[string] = proposals[:,ii]

    pickle.dump(to_pickle, open('%s/result_file_%s.p' %(
        log_folder, timestamp), 'wb'))

args = parse_arguments()

print(args)

timestamp = datetime.datetime.utcnow().strftime('%Y-%m-%d_%H-%M-%S')
print(timestamp)

if args.local == 1:
    log_folder = "Log_files"
else:
    log_folder = "/disk/scratch/REDACTED"

if args.folder is not None:
    log_folder += '/' + args.folder

Path(log_folder).mkdir(exist_ok=True)

log_file_str = '%s/run_file_%s.txt' %(log_folder, timestamp)
log_file = open(log_file_str, 'w')

seed = 13


# Input data settings
R = args.R 
D = 2
N = args.N

log_file.write('R: %s\tD: %s\tN: %s\tSeed: %s\n' %(R, D, N, seed))
if args.data_locs is not None:
    log_file.write('Data locations: %s\n min: \t%s, max \t%s' %(
        args.data_locs[0], args.data_locs[1], args.data_locs[2]))
    
# Inference settings
N_samp = args.N_samp # How many samples (of theta) to do 
ee_samp_ratio = args.M # How many times to sample eta per theta sample
burn_in = args.burn_in # Number of samples to ignore at start

log_file.write('N_samp: %s\tee_samp_ratio: %s\tburn_in: %s\n' %(
    N_samp, ee_samp_ratio, burn_in))
log_file.write(str(args))

print(args.kernel)

if args.kernel_hyp_file is not None:
    kernel, hyp_dict = pickle.load(open(args.kernel_hyp_file, 'rb'))
else:
    kernel, hyp_dict = None, None

prob = problems.Multiple_kernel_problem(
    args.kernel, parts=args.parts,
    hyp_dict=hyp_dict, kernel=kernel)

def build_eta_dict(args, constant, hd_helper=None, 
                   suffix='true'):
    eta_raw_dict = {}
    eta_constant = []
    
    if hd_helper is None:
        hd_helper = hd.Hyp_dict({'kernel_name': args.kernel}, 
                                parts=args.parts)
    eta_raw_dict['sigma_n'] = \
            problems.gamma_vars(
            shape=args.__dict__['sigma_n_'+suffix][0], 
            scale=args.__dict__['sigma_n_'+suffix][1])
    if 'sigma_n' in constant:
        eta_constant.append('sigma_n')
        
    if args.kernel in ['Combination', 'Temporal']:
        key_dict_list = hd_helper.key_dict
    else:
        key_dict_list = [hd_helper.key_dict]
    
    for kk in range(len(key_dict_list)):
        if len(key_dict_list) == 1:
            kk_str = ''
        else:
            kk_str = '%s_' %kk
        if type(key_dict_list[kk]) is dict:
            key_dict_sup = [key_dict_list[kk]]
        else:
            key_dict_sup = key_dict_list[kk]
        for jj in range(len(key_dict_sup)):
            if len(key_dict_sup) == 1:
                kk_jj = kk_str
            else:
                kk_jj = '%s%s_' %(kk_str, jj)
                
            for key in key_dict_sup[jj]:
                if key in ['sigma_rbf', 'sigma_hat']:
                    key = 'sigma_r'

                if key in constant:
                    eta_constant.append('%s%s' %(kk_jj, key))

                if key == 'direc':
                    eta_raw_dict['%sdir' %(kk_jj)] = \
                        problems.uniform_vars(low_b=0, high_b=np.pi)
                    if 'dir' in constant:
                        eta_constant.append('%sdir' %(kk_jj))

                elif key != 'sigma_2':
                    eta_raw_dict['%s%s' %(kk_jj, key)] = \
                        problems.gamma_vars(
                                shape=args.__dict__['%s_%s' %(
                                    key, suffix)][0],
                                scale=args.__dict__['%s_%s' %(
                                    key, suffix)][1])

                    args.__dict__['C%s%s_prop' %(kk_jj, key)] = \
                            args.__dict__['%s_prop' %key]
    print(args)
    return eta_raw_dict, eta_constant

eta_true_dict, ee_true_constant = build_eta_dict(
    args, suffix='true', constant=args.constant, hd_helper=hyp_dict)
print(ee_true_constant)
ee_true = prob.eta_vars(eta_true_dict, constant=ee_true_constant, args=args)
    
log_file.write('True values\n')
for key in eta_true_dict.keys():
    log_file.write('%s: \t%s\n' %(
        key, str(eta_true_dict[key])))

eta_init_dict, ee_init_constant = build_eta_dict(
    args, suffix='init', constant=args.constant, hd_helper=hyp_dict)

ee = prob.eta_vars(eta_init_dict, constant=ee_init_constant, args=args)

log_file.write('Initialisations\n')
for key in eta_true_dict.keys():
    log_file.write('%s: \t%s\n' %(
        key, str(eta_true_dict[key])))
    
# Draw theta from true distribution, and sample data | theta

np.random.seed(seed)

# Load or generate data
proposals = []
samples = []
locs = []
if args.data_locs is None:
    for rr in range(R):
        proposal = prob.sample_theta(ee_true)

        # NB: gives some samples at same location
        x_loc = np.random.randint(0, 28, (N, D))
        locs.append(x_loc)
        proposals.append(proposal.array())

        samp = np.random.multivariate_normal(
            np.zeros(N), prob.calc_K(x_loc, proposal=proposal))
        samples.append(samp)
        print(samp)
else:
    min_id = int(args.data_locs[1])
    max_id = int(args.data_locs[2])
    file_start = args.data_locs[0]
    R = max_id - min_id + 1
    for rr in range(min_id, max_id+1):
        try:
            plume_d = pickle.load(open(file_start+str(rr)+'.p', 'rb'))
            # Randomly choose N locations from the available samples
            idx = list(range(len(plume_d.domain)))
            np.random.shuffle(idx)
            idx = idx[:N]
            locs.append(plume_d.domain[idx, :])
            proposals.append(None)
            samples.append(plume_d.true_func[idx])
        except:
            print('File not found: %s' %(file_start+str(rr)+'.p'))
            R = R - 1


# Perform inference

start = time.time()

# Initiate theta variables
tt_vec = []
likelihood_tt_vec = []
for rr in range(R):
    tt = prob.sample_theta(ee)
    likelihood_tt = prob.calc_log_likelihood(
        tt, [locs[rr]], [samples[rr]])
    print(likelihood_tt)
    tt_vec.append(tt)
    likelihood_tt_vec.append(likelihood_tt)
tt = np.array(tt_vec)
likelihood_tt = np.array(likelihood_tt_vec)
    

# lists to collate samples
tt_samples = []
ee_samples = []

log_file.write('%s \t %s\n' %(-1, ee.array()))

# Do sampling
print('Starting sampling')
for ii in range(N_samp):
    try:
        tt, likelihood_tt = MCMC.resample_theta_multiple(
            tt, likelihood_tt, ee, locs, samples,
            prob.sample_theta, prob.calc_log_likelihood)
        for _ in range(ee_samp_ratio):
            ee, likelihood_ee = MCMC.resample_eta_by_parts(tt, ee,  
                log_pdf_theta_func=prob.pdf_theta_log_multiple)        
        tt_samples.append(tt)
        ee_samples.append(ee)
        log_file.write('%s \t %s \t %s\n' %(ii, likelihood_ee, ee.array()))
        if not args.local:
            print('%s \t %s \t %s\n' %(ii, likelihood_ee, ee.array()))
        if not ((likelihood_ee > -np.inf) and (likelihood_ee < np.inf)):
            log_file.write('Theta values:\n')
            log_file.write('jj, p(t|e), p(D|e) %s\n' %tt.order) 
            for jj in range(R):    
                log_file.write('%s \t %s \t %s \t %s\n' %(
                    jj, prob.pdf_theta_log(tt[jj], ee),
                prob.calc_log_likelihood(tt[jj], [locs[jj]], [samples[jj]]),
                    tt[jj].array()
                ))
                    
    except:
        print('An error occured!')
        pickle_state(ee_samples, tt_samples, args, samples, locs, proposals,
             timestamp, log_folder)

        dur = time.time() - start
        print('Duration: ', np.round(dur, 2))
        raise ValueError


    

pickle_state(ee_samples, tt_samples, args, samples, locs, proposals,
             timestamp, log_folder)
dur = time.time() - start
print('Duration: ', np.round(dur, 2))

log_file.write('Duration: %s\n' %np.round(dur, 2))

log_file.close()

print('Log file: ' + log_file_str)
